#!/usr/local/bin/ruby

# 1635, Mon 4 Jan 10 (NZDT)
#
# mk-hourly-stats.rb:  runs through telescope trace files
#   producing hourly statistics using newmon.rb
#
# For now, just write summary files into SummaryDir directory.
#   Will make an sql database from them one day.
#
# Copyright 2009-2011, Nevil Brownlee, CAIDA | U Auckland

require "#{File.dirname($0)}/owt-config"
Live = '/work/telescope/live'
SumFiletype = 'sum2'

require "#{Home}/owt-rb/owt-support" 

#require 'rubygems'
#require 'tzinfo'
#ServerCountry = 'America/Los_Angeles'  # Telescope timezone
#Ttz = TZInfo::Timezone.get(ServerCountry)

# Telescope times are UTC, but Ruby gives them to us in local time

def utc_gts(local_ts)  # Convert ts to gnuplot timefmt
  ds = local_ts.to_i+30  # strfmtime will round down to closest hour
#  begin
#    Time.at(Ttz.local_to_utc(ds).to_i).strftime("%Y%m%d.%H%M")
    Time.at(ds).gmtime.strftime("%Y%m%d.%H%M")
      # Assume that gmtime uses TZ environment variable
      # time format suitable for gnuplot !
#  rescue
#    "20091101.0100"  # Dummy to handle 'ambiguous time' at PDT changeover
#  end
end

if ARGV.size == 0
  print "No arguments given <<<\n";  exit
end

# Don't start another copy of mk-hourly-sum2
#   unless if has 'F' as its last command-line argument
#   Added Tue, 3 May 11 (NZST)

if ARGV[ARGV.size - 1] != 'F'
  $stdout.sync = true
  n_copies = 0
  IO.popen('ps uax | grep mk-hourly-sum2.rb') do |f|
    until f.eof?
      line = f.gets
      puts(line)
      n_copies += 1 if line =~ /ruby \.\/mk-hourly-sum2.rb/
    end
  end
  if n_copies != 1
    print "mk-hourly-sum2.rb already running, not starting another copy <<<\n"
    exit
  end
end

def set_start_date
  tn = Time.now.to_i 
  Time.at(tn - 4*24*3600).strftime('%Y-%m-%d') 
end

if ARGV.size == 0
  start_date = set_start_date()
elsif ARGV[0] == 'F'
  start_date = set_start_date()
else
  start_date = ARGV[0]
  last_date = ARGV[1]  # nil if only one arg
end
print "start_date = #{start_date}, last_date=#{last_date}.\n"

$break_loop = false
def sigtstp
  $stderr.print "  mk-hourly-sum2.rb will stop after current file\n"
  $break_loop = true
end 
trap(:TSTP, lambda {sigtstp})  # ^Z (Temporary stop)

trap(:INT, lambda {} )  # Ignore ^C (it's children will see the :INT)

Dir.glob("#{Live}/????-??-??").sort.each do |an|  # ? matches single char
  date = File.basename(an)
#  print "  date=#{date}, date<start_date=#{date < start_date}\n"
  next if date < start_date
  break if !last_date.nil? && date > last_date
  filetype = 'pcap.gz'
  sum_dir = "#{Home}/#{SummaryDir}"
  check_dir_exists(sum_dir)
  check_dir_exists("#{sum_dir}/#{date}")
  log_fn = "#{sum_dir}/#{date}/#{date}.fdf"

  Dir.glob("#{Live}/#{date}/*.#{filetype}").sort.each do |fulln|
    fn = File.basename(fulln)
    fn =~ /^net44\.(\d+)\.pcap.gz$/
    epoch = $1
    ts = utc_gts($1)

    trace_sz = sprintf("%.3f", File.size(fulln)/1000000000.0)  # GB
#    print "    fullfn=#{fulln}, fn=#{fn}, epoch=#{epoch}, ts=#{ts}, trace_sz=#{trace_sz}\n"

    sum_fn = "#{sum_dir}/#{date}/#{ts}.#{SumFiletype}"
    have_sum_file = already_there(sum_fn)
    print "sum_fn=#{sum_fn}, have_sum_file=#{have_sum_file}\n"

    if !already_there(sum_fn)
      cmd = "#{Home}/iatmon.rb pcapfile:#{fulln} #{sum_dir} #{log_fn} 0"
        # 0 => don't display Time or Stats messages
      print "cmd: #{cmd}\n"
      system(cmd)
      # exit  # Just do one file !!!
      exit if $break_loop
    end
  end
end
