#!/usr/bin/env ruby

# 1230, Fri 4 Jun 10 (NZST)
#
# mk-owt-config.rb: Makes 'default' Ruby config file
#   User will probably need to edit the owt-config file  <<<
#   so that it works in their local owtmon environment   <<<
#
# iatmon/owtmon: One-Way Traffic Monitor testbed
# Copyright (C) 2010-2011 by Nevil Brownlee, U Auckland | CAIDA | WAND
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>


require './owt-rb/monitorinfo'

h_name = `hostname`.chomp!
ha = h_name.split('.')
hostname = ha[0]

cfname = 'owt-config.rb'
old_config = false

if FileTest.exists?(cfname) && FileTest.size(cfname) != 0
  print "   owt-config file #{cfname} already exists, leaving it unchanged\n"
  old_config = true
else
  fname = "owt-config/#{hostname}.rb"
  if FileTest.exists?(fname) && FileTest.size(fname) != 0
    print "   owt-config file #{fname} exists in owt-config directory,\n"
    print "      copying it to #{cfname}\n"
    system("cp #{fname} #{cfname}")
    old_config = true
  end
end
if old_config
  cf = File.open(cfname, 'r')
  mi = MonitorInfo.new(cf)
  print "      Host (in owt-config.rb file) = #{mi.config[:Host]}\n"
  print "      *** this host is #{hostname} ***\n" \
    if mi.config[:Host] != hostname
  print "\n"
  exit
end

def p_exists?(fn)
  $LOAD_PATH.each {|path| return true if File.exist?(path + '/' + fn) }
  return true if File.exist?('./' + fn) 
  false
end

if p_exists?('IPprefix.bundle') || p_exists?('IPprefix.so')
  require 'IPprefix'
else
  print "Need IPprefix so we can generate a config file! <<<\n\n"
  exit
end

cf = File.new(cfname, 'w')
conf_hdr = <<END
# This file is a default version, generated by ./configure
#   >>> You probably need to edit it to make specify your   <<<
#   >>> local owtmon environment correctly                  <<< 

# Ruby owtmon programs use this config file by saying
#   require "#{File.dirname($0)}/owt-config"


# Default command-line parameters for owtmon . . .

SourceURI = 'pcapint:eth0'  # (libtrace-style) URI for input data source

SummaryDir = 'sum2'    # Directory where summary files are kept
                       #   (in YYYY-MM-DD subdirs) 

StatsFile = 'owt-stats'  # Statistics file for current day's summaries


# System constants . . .

SumFileType = 'sum2'   # Filetype for summary files

Hours_reqd = 1         # Summary file interval (hours)

MxLifetime = 172800    # Upper bound for lifetime distribs (seconds in 2 days)

N_Streams =   300000   # Memory allocation: number of one-way sources;
                       #    owtmon allocates blocks this size as needed

N_PktDatas = 4000000   # Memory allocation: number of packets for 
                       #   not-yet-analysed sources

MinPkts = 10           # Minimum number of packets to classify a source
                       #   (at end of a summary interval)

N_ProbePorts = 100     # Number of TCP and UDP probe ports kept in summaries

MeterLoc = 'Unknown'   # Normally use 3-letter airport code, e.g. SAN, AKL


# Network information . . .

END
cf.print conf_hdr

Home = Dir.pwd
cf.print "Home = '#{Home}'  # owtmon home directory\n\n"

hostname = `hostname`.chomp
Host = hostname.split('.')[0]
cf.print "Host = '#{Host}'\n\n"

require 'socket'
addr = Socket.gethostbyname(hostname)
af = addr[2];  abytes = addr[3]
LocalAddress = IPprefix.new(af == Socket::AF_INET6 ? 6 : 4, abytes,
  af == Socket::AF_INET6 ? 48 : 16)

cf.print "LocalAddress = '#{LocalAddress}'\n"
cf.print "      # This is the IP address prefix(es) for the network that\n"
cf.print "      #   your one-way packets' destination address fall within.\n"
cf.print "      # It may be a comma-separated list of mixed IPv4 and/or\n"
cf.print "      #   IPv6 addresses.  White space is ignored after a comma.\n"
cf.print "\n"

cf.close

print "   New config file #{cfname} generated,\n" +
      "      edit it for your local conditions\n\n"
