/* 1220, Sun  9 Dec 07 (NZDT)

   owt_mem.h:  Defines for owtmon memory management

   iatmon/owtmon: One-Way Traffic Monitor testbed
   Copyright (C) 2009-2011 by Nevil Brownlee, U Auckland | CAIDA | WAND

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/* #defines to declare variables and functions
   to allocate and manage stream_engine memory
   for struct 's_name'

   [ gcc -E to only run preprocessor
     gcc -c test.c
     gcc -o test -lpthread semaphore.o test.o ] */


/* (1) Declare global structs, usable in all .c files
       and templates for functions to work with them */

#define owt_mem_declare_globals(s_name) \
struct s_name##_space { \
   struct s_name *objects; \
   uint32_t mx_objects, initial_n_objects;  \
   int complained_about_objects, n_extra_blocks;  \
   struct s_name##_queue free_queue; \
   }; \
\
extern struct s_name##_space s_name##_space; \
void init_##s_name##_space( \
   struct s_name##_space *s_name##_space, int n_objects); \
void make_more_##s_name##_space( \
   struct s_name##_space *s_name##_space); \
struct s_name *get_##s_name##_object( \
   struct s_name##_space *s_name##_space); \
void free_##s_name##_object(struct s_name *this, \
   struct s_name##_space *s_name##_space);

/* (2) Define variables and functions (matching the templates) */

#define owt_mem_declare_functions(s_name) \
struct s_name##_space s_name##_space; \
\
void init_##s_name##_space( \
      struct s_name##_space *s_name##_space, int n_objects) { \
   int j; \
   s_name##_space->mx_objects =  /* Get object space */ \
      s_name##_space->initial_n_objects = n_objects; \
   if ((s_name##_space->objects = (struct s_name *)calloc( \
         n_objects, sizeof(struct s_name))) == NULL) \
      log_msg(LOG_ERR, 1, "Not enough memory for " #s_name " objects!"); \
   init_##s_name##_queue(&s_name##_space->free_queue); \
   for (j = 0; j != n_objects; ++j)  /* Initialise free queue */ \
      s_name##_push(&s_name##_space->free_queue, &s_name##_space->objects[j]); \
   s_name##_space->complained_about_objects = 0; \
   } \
\
void make_more_##s_name##_space( \
      struct s_name##_space *s_name##_space) { \
   struct s_name *new_space; \
   uint32_t j; \
   if ((new_space = (struct s_name *)calloc( \
         s_name##_space->initial_n_objects, sizeof(struct s_name))) == NULL) \
      log_msg(LOG_ERR, 1, "Not enough memory for more" #s_name " objects!"); \
   s_name##_q_lock(&s_name##_space->free_queue);   /* Add to free queue */ \
   s_name##_space->n_extra_blocks += 1; \
   for (j = 0; j != s_name##_space->initial_n_objects; ++j) \
      s_name##_push(&s_name##_space->free_queue, &new_space[j]); \
   s_name##_space->mx_objects += s_name##_space->initial_n_objects; \
   s_name##_q_unlock(&s_name##_space->free_queue); \
   } \
\
struct s_name *get_##s_name##_object( \
      struct s_name##_space *s_name##_space) { \
   struct s_name *this; \
   s_name##_q_lock(&s_name##_space->free_queue); \
   this = s_name##_shift(&s_name##_space->free_queue); \
   if (this == NULL) { \
      if (!s_name##_space->complained_about_objects) { \
         log_msg(LOG_WARNING, 0, "No free " #s_name " objects <<<"); \
         s_name##_space->complained_about_objects = 1; \
         } \
      s_name##_q_unlock(&s_name##_space->free_queue); \
      return NULL; \
      } \
   s_name##_q_unlock(&s_name##_space->free_queue); \
   return this; \
   } \
\
void free_##s_name##_object(struct s_name *this, \
      struct s_name##_space *s_name##_space) { \
   s_name##_q_lock(&s_name##_space->free_queue); \
   s_name##_push(&s_name##_space->free_queue, this); \
   s_name##_space->complained_about_objects = 0; \
   s_name##_q_unlock(&s_name##_space->free_queue); \
   }

/* (3) Define variables within a .c file */

#define owt_mem_declare_space(s_name) \
struct s_name##_space s_name##_space;
