/* 1541, Thu 17 Mar 11 (PDT)
   1628, Wed 22 Dec 10 (NZDT)
   1313, Wed 19 May 10 (NZST)
   1955, Thu 22 Apr 10 (PDT)
   1829, Sun  9 Aug 09 (NZST)

   owt_monitor.c: Ruby C interface to owtmon

   iatmon/owtmon: One-Way Traffic Monitor testbed
   Copyright (C) 2009-2011 by Nevil Brownlee, U Auckland | CAIDA | WAND

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "ruby.h"
#include "owt.h"

static VALUE IPprefix; /* IPprefix class */

static VALUE mSE;      /* OWTmonitor module */
static VALUE cST;      /* Monitor class */
static VALUE cStream;  /* OWTevent class */
static VALUE cSEmsg;   /* OWTmessage class */

#if 0 //xx
static struct opq_entry cs_oqe;

static VALUE cs_init(VALUE self, VALUE s_id) {
   rb_iv_set(self, "@id", s_id);
   return self;
   }

static VALUE cs_last_tsec(VALUE self) {
   struct stream *tsp;
   uint32_t ts;
   cs_oqe.s_id = FIX2INT(rb_iv_get(self, "@id"));
   tsp = stream_from_oqe(&cs_oqe);
   mutex_lock("cts", &stream_data_mutex);
   ts = (uint32_t)(tsp->last_us/1000000.0);
   mutex_unlock("cts", &stream_data_mutex);
   return INT2NUM(ts);
   }

static VALUE cs_ip_addresses(VALUE self) {
   struct stream *tsp;
   uint8_t *a;
   VALUE pa[3];
   VALUE addrs[2];  /* Result: [src_addr, dest_addr] */

   cs_oqe.s_id = FIX2INT(rb_iv_get(self, "@id"));
   tsp = stream_from_oqe(&cs_oqe);
   if (tsp->version == 4) {
      pa[0] = INT2FIX(4);  pa[2] = INT2FIX(32);
      a = tsp->src_addr.a.a;
      pa[1] = rb_str_new((char *)a, IP4_ADDR_LEN);
      addrs[0] = rb_class_new_instance(3, pa, IPprefix);
      }
   else {  /* v6 */
      pa[0] = INT2FIX(6);  pa[2] = INT2FIX(128);
      a = tsp->src_addr.a.a;
      pa[1] = rb_str_new((char *)a, IP6_ADDR_LEN);
      addrs[0] = rb_class_new_instance(3, pa, IPprefix);
      }
   pa[1] = Qnil;
   return rb_ary_new4(2, addrs);
   }

static VALUE cs_usage(VALUE self) {
   struct opq_entry cs_oqe;
   struct stream *sp;
   VALUE ra[4];  /* first_fwd, src_addr, tot_pkts, lifetime_us */
   VALUE pa[2];
   cs_oqe.s_id = FIX2INT(rb_iv_get(self, "@id"));
   sp = stream_from_oqe(&cs_oqe);

   pa[0] = INT2FIX(sp->version);
   if (sp->version == 4) {
      pa[1] = rb_str_new((char *)sp->src_addr.a.a, 4);
      }
   else pa[1] = rb_str_new((char *)sp->src_addr.a.a, 16);
   ra[0] = sp->first_fwd ? Qtrue : Qfalse;
   ra[1] = rb_class_new_instance(2, pa, IPprefix);
   ra[2] = INT2FIX(sp->tp + sp->fp);
   ra[3] = rb_float_new(sp->last_us - sp->first_us);
   return rb_ary_new4(4, ra);
   }

static VALUE cs_set_source_attribs(VALUE self, VALUE v_aa) {
   VALUE v_arg;
   struct stream *sp;

   cs_oqe.s_id = FIX2INT(rb_iv_get(self, "@id"));
   sp = stream_from_oqe(&cs_oqe);

   if (sp->state == SS_TWO_WAY)  /* ENOUGH -> TWO_WAY */
      return self;  /* Let it time out */

   if ((v_arg = rb_ary_shift(v_aa)) != Qnil) {  /* src_type */
      sp->src_type = NUM2INT(v_arg);
      }
   v_arg = rb_ary_shift(v_aa);  /* probe_port (may be nil) */
   if (v_arg == Qnil) sp->probe_port_valid = 0;
   else {
      sp->probe_port = NUM2INT(v_arg);  sp->probe_port_valid = 1;
      }
   if ((v_arg = rb_ary_shift(v_aa)) != Qnil)  /* mean_retry_time */
     sp->retry_time = NUM2DBL(v_arg);
   if ((v_arg = rb_ary_shift(v_aa)) != Qnil)  /* n_dest_ports */
      sp->n_dest_ports = NUM2INT(v_arg);
   if ((v_arg = rb_ary_shift(v_aa)) != Qnil)  /* pkt_len_mode */
      sp->pkt_len_mode = NUM2INT(v_arg);

   sp->analysed = 1;  sp->msg_queued = OQ_NULL;

#if MKSTATSCOUNTS  /* Now we know stream's src_type */
   sp->last_active_interval = stats_interval_nbr;
   active_sources_this_interval[sp->src_type] += 1;
   unique_hosts_this_interval[sp->src_type] += 1;
#endif
   return self;
   }
#endif  /* ---------------------- */

static VALUE msg_init(VALUE self, VALUE type, VALUE stream, VALUE text) {
   rb_iv_set(self, "@type", type);
   rb_iv_set(self, "@source", stream);
   rb_iv_set(self, "@text", text);
   return self;
   }

static char *interface_names[MXINTERFACES];  /* For start_meter() call */
static int if_names_sz, mx_events, mx_opq_entries,
   mx_streams, pw_q_sz, mx_stream_txts, mx_flow_watchers,
   mx_source_types;
   
static VALUE owtm_initialize(VALUE self, VALUE v_params) {
   /* Just call class init() method, i.e. owtm_init() */
   rb_funcall(self, rb_intern("init"), 1, v_params);
   return self;
   }

static VALUE owtm_init(int argc, VALUE *argv, VALUE self) {
   VALUE v_interfaces, v_live,  v_streams;
   VALUE v_ra, v_arg, rs_if;
   int rs, if_len, l_len, j;
   char *if_name, *l_name;

   uid_from_cwd();  /* Set my_uid and se_uid (from ~/ directory) */

   logname = (char *)"owt.log";  /* Default log name */
   mx_events = DFMXEVENTS;  /* Object memory sizes */
   mx_opq_entries = DFMXOPQENTRIES;
   pw_q_sz = DFMXPWQUEUESZ;
   mx_stream_txts = DFMXSTREAMTXTS;
   mx_flow_watchers = DFMXFLOWWATCHERS;
   n_local_addrs = 0;

   traces_to_process = DFTRACESTOPROCESS;

   rs = rb_scan_args(argc, argv, "3*",  /* Get args (from Streamer.new) */
      &v_interfaces, &v_live, &v_streams,
      &v_ra);
   if_names_sz = (int)RARRAY_LEN(v_interfaces);
   if (if_names_sz > MXINTERFACES)
      log_msg(LOG_ERR, 27, (char *)"More than %d interfaces specified !!!");
   for (j = 0; j != if_names_sz; ++j) {
      rs_if = rb_ary_shift(v_interfaces);
      if_name = RSTRING_PTR(rs_if);  if_len = (int)RSTRING_LEN(rs_if);
      interface_names[j] = malloc(if_len+1);
      strncpy(interface_names[j], if_name, if_len+1);
      }
   live_source = v_live == Qtrue;
   mx_streams = NUM2INT(v_streams);
   // mx_opq_entries = mx_streams;  /* 6 Aug 10 */

   if ((v_arg = rb_ary_shift(v_ra)) != Qnil)
     time_rec_interval = NUM2INT(v_arg);
   if ((v_arg = rb_ary_shift(v_ra)) != Qnil)
     min_pkts_to_classify = NUM2INT(v_arg);
   if ((v_arg = rb_ary_shift(v_ra)) != Qnil)
      mx_source_types = NUM2INT(v_arg);
   if ((v_arg = rb_ary_shift(v_ra)) != Qnil)
      traces_to_process = NUM2INT(v_arg);
   if ((v_arg = rb_ary_shift(v_ra)) != Qnil) {  /* 17 Mar 11 */
      char *location;  int location_len;
      location = RSTRING_PTR(v_arg);  location_len = (int)RSTRING_LEN(v_arg);
      meter_loc = malloc(location_len+1);
      strncpy(meter_loc, location, location_len+1);
      }
   if ((v_arg = rb_ary_shift(v_ra)) != Qnil) {  /* 27 May 11 */
     l_name = RSTRING_PTR(v_arg);  l_len = (int)RSTRING_LEN(v_arg);
      logname = malloc(l_len+1);
      strncpy(logname, l_name, l_len+1);
      }
   if ((v_arg = rb_ary_shift(v_ra)) != Qnil) {  /* 23 Mar 11 */
      save_distributions = NUM2INT(v_arg);
      }

   if ((log_file = fopen(logname,"a")) == NULL) {
      fprintf(stderr,"Failed to open logfile '%s':%s\n",
         logname, strerror(errno));
      exit(321);  /* Don't try to run without log file! */
      }
#if 0
   printf("streams=%u, timerecinterval=%u, minpkts=%u, types=%u, hours=%u, log=%s\n",
      mx_streams, time_rec_interval, min_pkts_to_classify,
	  mx_source_types, traces_to_process, logname);
#endif
   return self;
   }

static VALUE owtm_info(VALUE self) {
   VALUE ra[2];
   ra[0] = rb_str_new2(METER_VERSION);
#if STREAMTIMEOUTS
   ra[1] = Qtrue;
#else
   ra[1] = Qfalse;
#endif
   return rb_ary_new4(sizeof(ra)/sizeof(VALUE), ra);
   }

struct key_val *get_kv(struct hashtable *h, int *length);
struct key_val *kv_dst_ports, *kv_tcp_keys;

static VALUE r_array(uint32_t *ca, int len) {
   VALUE ra[256];  int j;
   for (j = 0; j != len; ++j)
      ra[j] = UINT2NUM(ca[j]);
   return rb_ary_new4(len, ra);
   }

static void r_kv(VALUE *r_keys, VALUE *r_counts, struct hashtable *h) {
   struct key_val *kv;  int len, j;
   VALUE *rkey, *rcount;
   kv = get_kv(h, &len);
   if (len == 0) {
      *r_keys = *r_counts = Qnil;
      return;
      }
   rkey = (VALUE *)malloc(len*sizeof(VALUE));
   rcount = (VALUE *)malloc(len*sizeof(VALUE));
   for (j = 0; j != len; ++j) {
     rkey[j] = UINT2NUM(kv[j].key);
     rcount[j] = UINT2NUM(kv[j].count);
     }
   *r_keys = rb_ary_new4(len, rkey);
   *r_counts = rb_ary_new4(len, rcount);
   free(kv);
   free(rkey);  free(rcount);
   }

static VALUE owtm_unanalysed(VALUE self) {
   VALUE ra[12];
   ra[0] = UINT2NUM(n_unanalysed_sources);
   ra[1] = UINT2NUM(n_unanalysed_pkts);

   if (n_unanalysed_sources == 0)
      ra[2] = ra[3] = ra[4] = ra[5] = ra[6] = ra[7] =
         ra[8] = ra[9] = ra[10] = ra[11] = Qnil;
   else {
      ra[2] = r_array(ua_pkts, MN_PKTS_FOR_DISCARD);
      ra[3] = r_array(ua_src_addr, 256);
      ra[4] = r_array(ua_dst_addr, 256);
      ra[5] = r_array(ua_ttl, 256);
      ra[6] = r_array(ua_proto, 256);
      ra[7] = UINT2NUM(ua_gt1prot);
      r_kv(&ra[8],&ra[9], ua_dst_ports);
      r_kv(&ra[10],&ra[11], ua_tcp_keys);
      n_unanalysed_sources = 0;
      }
   return rb_ary_new4(sizeof(ra)/sizeof(VALUE), ra);
   }

#if HALF_WAY_STREAMS
static void addr_r_kv(VALUE *r_keys, VALUE *r_counts, struct hashtable *h) {
   struct addr_key_val *kv;  int len, j;
   VALUE *raddr, *rcount, pa[3];
   uint8_t *a;
   kv = addr_get_kv(h, &len);  /* Finished with ht, desgtroys it */
   if (len == 0) {
      *r_keys = *r_counts = Qnil;
      return;
      }
   raddr = (VALUE *)malloc(len*sizeof(VALUE));
   rcount = (VALUE *)malloc(len*sizeof(VALUE));
   for (j = 0; j != len; ++j) {
      pa[0] = INT2FIX(4);  pa[2] = Qnil;
      a = kv[j].key.a.a;
      pa[1] = rb_str_new((char *)a, IP4_ADDR_LEN);
      raddr[j] = rb_class_new_instance(3, pa, IPprefix);
      rcount[j] = UINT2NUM(kv[j].count);
      }
   *r_keys = rb_ary_new4(len, raddr);
   *r_counts = rb_ary_new4(len, rcount);
   free(kv);
   free(raddr);  free(rcount);
   }
#endif

static VALUE owtm_log_msg(VALUE self, VALUE rdie, VALUE rmsg) {
   log_msg(LOG_WARNING, NUM2INT(rdie), RSTRING_PTR(rmsg));
   return self;
   }

static VALUE owtm_new_log_name(VALUE self, VALUE name) {
   char *l_name;  int l_len;
   fclose(log_file);  free(logname);
   l_name = RSTRING_PTR(name);  l_len = (int)RSTRING_LEN(name);
   logname = malloc(l_len+1);
   strncpy(logname, l_name, l_len+1);
    if ((log_file = fopen(logname,"a")) == NULL) {
      fprintf(stderr,"Failed to open new logfile '%s':%s\n",
         logname, strerror(errno));
      exit(321);  /* Don't try to run without log file! */
      }
   return self;
   }

static uint32_t ipv4_dest_addr_b2(uint32_t a) {
   return (ntohl(a) >> 16) & 0x000000FF;  /* 2nd byte */
   }

static uint32_t ipv4_dest_addr_b3(uint32_t a) {
   return (ntohl(a) >> 8) & 0x000000FF;  /* 3nd byte */
   }

static VALUE owtm_set_local_addrs(VALUE self, VALUE v_addr) {
   VALUE v_prefix;
   int v, nb, len;
   uint8_t *ap;
   for (n_local_addrs = 0;;) {
      v_prefix = rb_ary_shift(v_addr);
      if (v_prefix == Qnil) 
	 return INT2FIX(n_local_addrs);
      if (n_local_addrs == MX_LOCAL_ADDRS) rb_raise(rb_eArgError,
         "set_local_addrs(): MX_LOCAL_ADDRS is %d", MX_LOCAL_ADDRS);
      v =  FIX2INT(rb_iv_get(v_prefix, "@version"));
      nb = v == 4 ? IP4_ADDR_LEN : IP6_ADDR_LEN;
      ap = (uint8_t *)RSTRING_PTR(rb_iv_get(v_prefix, "@addr"));
      len = FIX2INT(rb_iv_get(v_prefix, "@length"));
      memcpy(la[n_local_addrs].a.a, ap, nb);
      la[n_local_addrs].len = len;
      la[n_local_addrs].ver = v;
      if (v == 4) {
         if (len == 8) dest_addr_bin = ipv4_dest_addr_b2;
         else if (len == 16) dest_addr_bin = ipv4_dest_addr_b3;
         }
      n_local_addrs += 1;
      }
   return self;
   }

static VALUE owtm_set_uris(VALUE self, VALUE u_a) {
   other_uris = (char **)malloc((traces_to_process-1)*sizeof(char *));
   VALUE u_s;  uint32_t j;
   for (j = 0;  j != traces_to_process-1; ++j) {
      u_s = rb_ary_shift(u_a);
      other_uris[j] = malloc(RSTRING_LEN(u_s)+2);
      strcpy(other_uris[j], RSTRING_PTR(u_s));
      }
   next_uri_nbr = 0;
   return self;
   }

static VALUE owtm_open(VALUE self) {
   start_meter(interface_names, if_names_sz, mx_events,
      mx_opq_entries, mx_streams,
      pw_q_sz, mx_stream_txts, mx_flow_watchers,
      mx_source_types, traces_to_process);
   return UINT2NUM(start_ts);
   }

static VALUE owtm_close(VALUE self) {
   request_shutdown = 1;
   printf("owtm_close() called\n");  fflush(stdout);  //##
   return self;
   }

#if 0
static VALUE owtm_nodename(VALUE self) {
   return rb_str_new2(node.nodename);
   }
#endif

static VALUE owtm_each(VALUE self) {
   struct opq_entry *oqe;
   VALUE attribs[3];  /* SEmessage: @type, @id, @text */
   VALUE sa[1];       /* Stream: @id */

   for (;;) {
      t_semaphore_down(&opq_ready);  /* Wait for a full block */

      opq_entry_q_lock(&opq);
      oqe = opq_entry_shift(&opq);
      opq_entry_q_unlock(&opq);
      if (oqe->type == OQ_TXTMSG) {
         attribs[0] = INT2FIX(OQ_TXTMSG);
         attribs[1] = INT2NUM(oqe->ts);
         attribs[2] = rb_str_new2(oqe->msg);
         rb_yield(rb_class_new_instance(3, attribs, cSEmsg));

         free_stream_txt_object(
            (struct stream_txt *)oqe->msg, &stream_txt_space);
         if (strncmp(oqe->msg, "#Shut", 5) == 0)
	    exit(7);  /* Ruby code should have closed fdf */
         }
      else {
   	 sa[0] = INT2FIX(oqe->s_id);  /* Stream, only has @id so far */
         attribs[0] = INT2FIX(oqe->type);
	 attribs[1] = rb_class_new_instance(1, sa, cStream);
         attribs[2] = Qnil;
         rb_yield(rb_class_new_instance(3, attribs, cSEmsg));
         }
      free_opq_entry_object(oqe, &opq_entry_space);

      t_semaphore_up(&opq_empty);  /* Signal an empty block */
      }
   return self;
   }

static int get_stream_sizes(struct stream *sp) {
   rb_yield(INT2FIX(sp->tp));
   return 0;
   }

static uint32_t n_streams_written;
static FILE *df = NULL;  /* For IAT distribution dumps */

static VALUE owtm_each_stream_size(VALUE self) {
   /* CAUTION: locks stream table during walk */
   struct flow_watcher_control *fw;
   struct stream_queue *sq;
   int nfw, k;

   mutex_lock("ess", &stream_data_mutex);
   nfw = flow_watcher_control_space.mx_objects;
   for (k = 0; k != nfw; ++k) {
      fw = &flow_watcher_control_space.objects[k];
      sq = &fw->sq;
      stream_walk_up(sq, get_stream_sizes);
      }
   mutex_unlock("ess", &stream_data_mutex);
   return self;
   }

#if STREAMTIMEOUTS
static int deleted_inactive;
static void delete_stream(struct stream *sp) {
   sp->state = SS_DELETE;  /* Let watcher time it out */
   sp->timeout = STO_ONE_WAY;
   }
#endif
static uint32_t two_way_streams;

#if ERF_DIRECTION
void update_direction_counts(struct stream *sp) {
   if (sp->state == SS_TWO_WAY) {
      if (sp->direction == 1) two_way_streams_1 += 1;
      else if (sp->direction == 2) two_way_streams_2 += 1;
      else two_way_streams_3 += 1;
      }
   else if (sp->state == SS_ONE_WAY) {
      if (sp->direction == 1) one_way_streams_1 += 1;
      else if (sp->direction == 2) one_way_streams_2 += 1;
      else one_way_streams_3 += 1;
      }
   else {
      if (sp->direction == 1) half_way_streams_1 += 1;
      else if (sp->direction == 2) half_way_streams_2 += 1;
      else half_way_streams_3 += 1;
      }
   }
#endif

static int get_stream_attribs(struct stream *sp)
{  /* Return 0 to continue, 1 to stop walking */

#if ERF_DIRECTION
   update_direction_counts(sp);
#endif

   if (sp->state != SS_ONE_WAY) {
      update_summary_info(sp);
      sp->tp = sp->fp = 0;  sp->tb = sp->fb = 0;
      return 0; 
      }
   if (sp->state == SS_DELETE) return 0;

#if STREAMTIMEOUTS
   if (sp->active_since_last_sum == 0) {  /* Inactive this interval */
      if (sp->inactive_last_sum == 1) {
         delete_stream(sp);  /* Inactive at least one whole #Record interval */
         deleted_inactive += 1;
         return 0;
         }
      sp->inactive_last_sum = 1;  /* Inactive at last #Record */
      return 0;
      }
   sp->inactive_last_sum = 0;  /* Active this interval */
   sp->active_since_last_sum = 0;
#endif

   if (sp->version == 4) {
      v4_one_way_streams += 1;
      v4_one_way_pkts += sp->tp;  v4_one_way_bytes += sp->tb;
      }
   else {
      v6_one_way_streams += 1;
      v6_one_way_pkts += sp->tp;  v6_one_way_bytes += sp->tb;
      }

   if (sp->flow_info.fi4 == NULL) {  /* No flow_info */
      update_ua_stats(sp);  /* Un-analysed source info */
      return 0;  /* No flow_info (or distrib) struct */
      }

   if (sp->last_us-sp->first_us < MIN_LIFETIME)
      return 0;  /* Too short for sensible statistics */

   int src_type = analyse_flow(sp);  /* Sets probe_port globals */
   int src_group;
   if (sp->flow_info.fi4->d != NULL)
      src_group = compute_distrib_stats(sp);
   else src_group = 0;

   if (save_distributions) {
      struct distribution *d = sp->version == 4 ?
         sp->flow_info.fi4->d : sp->flow_info.fi6->d;

      if (save_distributions == 2) {  /* 'D' parameter */
         if (df == NULL) {  /* First source distrib to be written */
            n_streams_written = 0;
            df = fopen("iat-distribs", "w");
            write_distrib_info(df, sp, d);
            }
#if STREAMS_TO_WRITE
         if (n_streams_written < STREAMS_TO_WRITE) {
            dump_source_info(df, sp, src_type, d);
            n_streams_written += 1;
            }
#else
         dump_source_info(df, sp, src_type, d);
#endif
         }
      else {
         if (df == NULL) {  /* First source distrib to be written */
            n_streams_written = 0;
            df = fopen("types+groups", "w");
            fprintf(df, "#trace: %s\n", trace_name);
            fprintf(df, "#starttime: %s (UTC)\n", fmt_time(1, &start_ts));
            fprintf(df, "#meter: %s\n", meter_loc);
            fprintf(df, "#attributes: type group  seq_nbr  src_addr\n");
            }
         fprintf(df, "%02d %02d  %06d  %s\n",
            src_type,src_group, n_streams_written, key_to_s(sp));
         n_streams_written += 1;
         }
      }

   VALUE ra[13];
      /* src_type probe_port to_pkts  total_lifetime tot_count  version
            too_few_pkts protocol  src_group  flags_mode keys_mode
            v6_addrs_s32 to_bytes */

   ra[0] = INT2FIX(src_type);
   ra[1] = probe_port_valid ? INT2FIX(probe_port) : Qnil;
   ra[2] = INT2FIX(sp->tp);
   ra[3] = rb_float_new(sp->last_us - sp->first_us);
   ra[4] = ULL2NUM(sp->totp);
   ra[5] = INT2FIX(sp->version);
   ra[6] = too_few_pkts ? Qtrue : Qfalse;
   ra[7] = INT2FIX(protocol);
   ra[8] = INT2FIX(src_group);
   ra[9] = INT2FIX(flags_mode);
   ra[10] = INT2FIX(keys_mode);
   if (sp->version == 4) {
      ra[11] = Qnil;
      }
   else { 
      ra[11] = UINT2NUM(ntohl(sp->src_addr.a.v4));
      }
#if STREAMTIMEOUTS
   init_flow_info(sp);  /* Zero flow_info counts */
#endif
   ra[12] = ULL2NUM(sp->tb);
   sp->tp = 0;  sp->tb = 0;

   rb_yield(rb_ary_new4(sizeof(ra)/sizeof(VALUE), ra));
   return 0;
   }

static VALUE owtm_each_stream(VALUE self) {  /* pkt_datas for src_host self */
   struct flow_watcher_control *fw;
   struct stream_queue *sq;
   int nfw, k;
#if STREAMTIMEOUTS
   char msg[120];  /* For #Streams message */
#endif
   VALUE r = Qnil;

   printf("*** Starting owtm_each_stream()\n");  // <<<<<<
   two_way_streams = 0;   
   nfw = flow_watcher_control_space.mx_objects;
#if STREAMTIMEOUTS
   for (;;) {  /* Wait for CV_IDLE */
      pthread_mutex_lock(&count_mutex);
      if (CV_IDLE) break;  /* Leave cv locked */
      else if (CV_CHECKING)  /* Must wait for timeouts to finish */
         pthread_cond_wait(&count_cv, &count_mutex);
      else log_msg(LOG_ERR, 1, (char *)"Attempt to Record: while CV_RECORDING <<<");
      pthread_mutex_unlock(&count_mutex);  /* Let watchers change count */
      }
   count -= 1;  /* Set count = CV_RECORDING */
   pthread_mutex_unlock(&count_mutex);
   deleted_inactive = 0;  //dd
#endif
   for (k = 0; k != nfw; ++k) {  /* Record: processing */
      fw = &flow_watcher_control_space.objects[k];
      sq = &fw->sq;
      stream_q_lock(sq);
      stream_walk_up(sq, get_stream_attribs);
      stream_q_unlock(sq);
      }
   if (save_distributions) fclose(df);
#if STREAMTIMEOUTS
   pthread_mutex_lock(&count_mutex);
   count += 1;  /* Set count = CV_IDLE */
   pthread_mutex_unlock(&count_mutex);
   sprintf(msg ,"#Streams: two=%u, v4=%llu, v6=%llu, inactive=%d",
      two_way_streams, v4_one_way_streams, v6_one_way_streams,
      deleted_inactive);
   r = rb_str_new2(msg);
#endif
   return r;
   }

#if 0  //??
static int get_stream_info(struct stream *sp)
{  /* Return 0 to continue, 1 to stop walking */
      /* src_type  to_pkts  first_us last_us  probe_port  src_addr */
  VALUE ra[6], pa[3];
  uint8_t *a;

   ra[0] = INT2FIX(sp->src_type);
   ra[1] = INT2FIX(sp->tp);
   ra[2] = rb_float_new(sp->first_us);
   ra[3] = rb_float_new(sp->last_us);
   ra[4] = sp->probe_port_valid ? INT2FIX(sp->probe_port) : Qnil;
   if (sp->version == 4) {
      pa[0] = INT2FIX(4);  pa[2] = Qnil;
      a = sp->src_addr.a.a;
      pa[1] = rb_str_new((char *)a, IP4_ADDR_LEN);
      ra[5] = rb_class_new_instance(3, pa, IPprefix);
      }
   else {  /* v6 */
      pa[0] = INT2FIX(6);  pa[2] = Qnil;
      a = sp->src_addr.a.a;
      pa[1] = rb_str_new((char *)a, IP6_ADDR_LEN);
      ra[5] = rb_class_new_instance(3, pa, IPprefix);
      }

   rb_yield(rb_ary_new4(sizeof(ra)/sizeof(VALUE), ra));
   return 0;
   }

static VALUE owtm_each_str_info(VALUE self) {  /* for owt-sti.rb */
   struct flow_watcher_control *fw;
   struct stream_queue *sq;
   int nfw, k;

   nfw = flow_watcher_control_space.mx_objects;
   for (k = 0; k != nfw; ++k) {
      fw = &flow_watcher_control_space.objects[k];
      sq = &fw->sq;
      stream_walk_up(sq, get_stream_info);
      }
   return self;
   }

static VALUE owtm_interval_stats(VALUE self) {
#if MKSTATSCOUNTS
   VALUE ra[2];
   VALUE as[mx_source_types], uh[mx_source_types];
   //x      ps[mx_source_types];
   int st, m;
   for (st = 0; st != mx_source_types; ++st) {
      VALUE active_sources[STATS_INTERVALS_PER_HOUR+1],
         /* +1 because we see a few packets recorded after the hour change! */
         /*   Also, type 0 sources are recorded only at EOF */
	//x	 pkts_seen[STATS_INTERVALS_PER_HOUR+1],
	 unique_hosts[STATS_INTERVALS_PER_HOUR+1];
      for (m = 0; m != STATS_INTERVALS_PER_HOUR+1; ++m) {
        active_sources[m] = UINT2NUM(active_sources_per_stats_interval[m][st]);
        unique_hosts[m] = UINT2NUM(unique_hosts_per_stats_interval[m][st]);
	//x        pkts_seen[m] = UINT2NUM(pkts_seen_per_stats_interval[m][st]);
        }
      as[st] = rb_ary_new4(STATS_INTERVALS_PER_HOUR+1, active_sources);
      uh[st] = rb_ary_new4(STATS_INTERVALS_PER_HOUR+1, unique_hosts);
      //x      ps[st] = rb_ary_new4(STATS_INTERVALS_PER_HOUR+1, pkts_seen);
      }
   ra[0] = rb_ary_new4(mx_source_types, as);
   ra[1] = rb_ary_new4(mx_source_types, uh);
   //x   ra[2] = rb_ary_new4(mx_source_types, ps);
   return rb_ary_new4(2, ra);
#else
   return Qnil;
#endif
   }
#endif

static VALUE owtm_c_addr_distribs(VALUE self) {
   VALUE counts[27];
   counts[0] = ULL2NUM(v4_one_way_pkts);
   counts[1] = ULL2NUM(v4_one_way_bytes);
   counts[2] = ULL2NUM(v4_two_way_pkts);
   counts[3] = ULL2NUM(v4_two_way_bytes);
   counts[4] = ULL2NUM(v4_one_way_streams);
   counts[5] = ULL2NUM(v4_two_way_streams);  
   counts[6] = ULL2NUM(v6_one_way_pkts);
   counts[7] = ULL2NUM(v6_one_way_bytes);
   counts[8] = ULL2NUM(v6_two_way_pkts);
   counts[9] = ULL2NUM(v6_two_way_bytes);
   counts[10] = ULL2NUM(v6_one_way_streams);
   counts[11] = ULL2NUM(v6_two_way_streams);
#if HALF_WAY_STREAMS
   counts[12] = ULL2NUM(v4_half_way_bytes);
   counts[13] = ULL2NUM(v4_half_way_pkts);
   counts[14] = ULL2NUM(v4_half_way_streams);
   counts[15] = ULL2NUM(v6_half_way_bytes);
   counts[16] = ULL2NUM(v6_half_way_pkts);
   counts[17] = ULL2NUM(v6_half_way_streams);
#else
   counts[12] = counts[13] = counts[14] = 
      counts[15] = counts[16] = counts[17] = Qnil;
#endif
#if ERF_DIRECTION
   counts[18] = ULL2NUM(one_way_streams_1);
   counts[19] = ULL2NUM(one_way_streams_2);
   counts[20] = ULL2NUM(one_way_streams_3);
   counts[21] = ULL2NUM(two_way_streams_1);
   counts[22] = ULL2NUM(two_way_streams_2);
   counts[23] = ULL2NUM(two_way_streams_3);
   counts[24] = ULL2NUM(half_way_streams_1);
   counts[25] = ULL2NUM(half_way_streams_2);
   counts[26] = ULL2NUM(half_way_streams_3);
#else
   counts[18] = counts[19] = counts[20] = 
      counts[21] = counts[22] = counts[23] =
      counts[24] = counts[25] = counts[26] = Qnil;
#endif
   VALUE ra[4];
   ra[0] = rb_ary_new4(sizeof(counts)/sizeof(VALUE), counts);
   ra[1] = ra[2] = ra[3] = Qnil;
#if HALF_WAY_STREAMS
   addr_r_kv(&ra[2],&ra[3], half_src_addrs);
#endif
#if MKDESTCOUNTS
   VALUE val[NDESTBINS];
   VALUE das[mx_source_types];
   int st, bin;
   for (st = 0; st != mx_source_types; ++st) {
      for (bin = 0; bin != NDESTBINS; ++bin) {
 	 val[bin] = UINT2NUM(dest_addrs_per_source_type[st][bin]);
         }
      das[st] = rb_ary_new4(NDESTBINS, val);
      }
   ra[1] = rb_ary_new4(mx_source_types, das);
#endif
   reset_summary_info();
   return rb_ary_new4(sizeof(ra)/sizeof(VALUE), ra);
   }

#if 0  /* Gives Ruby GC problems when called from C thread! */
static VALUE glob_next_uri(VALUE self, VALUE fnv) {
   VALUE fn_rv =  /* global functions live in Kernel class */
      rb_funcall(rb_mKernel, rb_intern("next_uri"), 1, fnv);
   return fn_rv;
   }

char *c_next_uri(char *uri) {
   VALUE uri_v = rb_str_new2(uri);
   VALUE new_uri_v =  /* global functions live in Kernel class */
      rb_funcall(rb_mKernel, rb_intern("next_uri"), 1, uri_v);
   char *ustr = RSTRING_PTR(new_uri_v);
   int ulen = RSTRING_LEN(new_uri_v);
   char *rs = malloc(ulen+2);
   strcpy(rs, ustr);
   return rs;
   }
#endif

void Init_OWTmonitor() {
   /* rb_require("IPprefix");
         So that we can run in owt-cron-env environment, owtmon.rb
         must require IPprefix.so  before  require OWTmonitor */
   IPprefix = rb_path2class("IPprefix");

   mSE = rb_define_module("OWTmonitor");

   rb_define_global_const("OQ_TXTMSG",  INT2FIX(OQ_TXTMSG));
   rb_define_global_const("OQ_TIMED_OUT", INT2FIX(OQ_TIMED_OUT));
   rb_define_global_const("OQ_END_OF_FILE", INT2FIX(OQ_END_OF_FILE));

#if 0 //xx
   cStream = rb_define_class("OWTevent", rb_cObject);
   memset(&cs_oqe, 0, sizeof(struct opq_entry));
   rb_define_method(cStream, "initialize", cs_init, 1);
   rb_define_method(cStream, "last_tsec", cs_last_tsec, 0);
   rb_define_method(cStream, "usage", cs_usage, 0);
   rb_define_method(cStream, "set_source_attribs", cs_set_source_attribs, 1);
#endif
   cST = rb_define_class_under(mSE, "Monitor", rb_cObject);
   rb_define_method(cST, "initialize", owtm_initialize, 1);
   rb_define_method(cST, "owt_init", owtm_init, -1);
   rb_define_method(cST, "info", owtm_info, 0);
   rb_define_method(cST, "unanalysed", owtm_unanalysed, 0);
   rb_define_method(cST, "log_msg", owtm_log_msg, 2);
   rb_define_method(cST, "new_log_file", owtm_new_log_name, 1);
   rb_define_method(cST, "open", owtm_open, 0);
   rb_define_method(cST, "set_uris", owtm_set_uris, 1);
   rb_define_method(cST, "set_local_addrs", owtm_set_local_addrs, 1);
   rb_define_method(cST, "close", owtm_close, 0);
   //??   rb_define_method(cST, "nodename", owtm_nodename, 0);
   rb_define_method(cST, "each", owtm_each, 0);
   rb_define_method(cST, "each_stream_size", owtm_each_stream_size, 0);
   rb_define_method(cST, "each_stream", owtm_each_stream, 0);
   //??   rb_define_method(cST, "each_str_info", owtm_each_str_info, 0);
   //??   rb_define_method(cST, "interval_stats", owtm_interval_stats, 0);
   rb_define_method(cST, "c_addr_distribs", owtm_c_addr_distribs, 0);

   cSEmsg = rb_define_class("OWTmessage", rb_cObject);
   rb_define_method(cSEmsg, "initialize", msg_init, 3);
   }
