/* 1541, Thu 17 Mar 11 (PDT)
   1700, Sat 17 Dec 10 (NZDT)
   1955, Thu 22 Apr 10 (PDT)
   1445, Wed  5 Dec 07 (NZDT)

   owt_support.c:  Support routines for owtmon

   iatmon/owtmon: One-Way Traffic Monitor testbed
   Copyright (C) 2009-2011 by Nevil Brownlee, U Auckland | CAIDA | WAND

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "owt.h"

/* Define global variables (declared extern in owt.h) */

char *logname;
FILE *log_file;
struct stream *extra_streams[MX_EXTRA_BLOCKS+1];
int extra_stream_incr, n_extra_streams;
int save_distributions;  /* Non-zero to print source distribs */

#if MKSTATSCOUNTS
uint32_t  /* Per-interval stats arrays for each source type .. */
//x   **pkts_seen_per_stats_interval, *pkts_seen_this_interval,
   **active_sources_per_stats_interval, *active_sources_this_interval,
   **unique_hosts_per_stats_interval, *unique_hosts_this_interval;
uint32_t stats_interval_nbr;
#endif

#if MKDESTCOUNTS
uint32_t **dest_addrs_per_source_type;
#define src_addr_bin(a)  ((ntohl(a) >> 24) & 0x000000FF)  /* 1st byte */
uint32_t (*dest_addr_bin)(uint32_t a);
#endif

struct ip_address la[MX_LOCAL_ADDRS];
int n_local_addrs;

uint32_t n_source_types, pkts_to_record, min_pkts_to_classify,
   traces_to_process;
char **other_uris;
int next_uri_nbr;

int time_rec_interval, intervals_this_sum_file;

int n_lost_stream_v4s, n_lost_stream_v6s;
uint64_t last_drops;

uint32_t mx_active_streams, mx_active_opq_entries,
   mx_active_v4_flow_infos, mx_active_v6_flow_infos,
   av_cl_count, n_hc_searches,  src_av_cl_count, src_n_hc_searches;
uint32_t n_not_ip,n_frags,n_too_small,n_no_local_address,
   n_v6_opaque;

#if HALF_WAY_STREAMS
struct hashtable *half_src_addrs;
#endif

uint64_t  /* Overall stats */
   v4_one_way_bytes,  v4_one_way_pkts,  v4_one_way_streams,
   v6_one_way_bytes,  v6_one_way_pkts,  v6_one_way_streams,
   v4_two_way_bytes,  v4_two_way_pkts,  v4_two_way_streams,
   v6_two_way_bytes,  v6_two_way_pkts,  v6_two_way_streams,
   v4_half_way_bytes, v4_half_way_pkts, v4_half_way_streams,
   v6_half_way_bytes, v6_half_way_pkts, v6_half_way_streams;

#if ERF_DIRECTION
 uint64_t
   one_way_streams_1, one_way_streams_2, one_way_streams_3,
   two_way_streams_1, two_way_streams_2, two_way_streams_3, 
   half_way_streams_1, half_way_streams_2, half_way_streams_3;
#endif

uint32_t start_ts,  /* Seconds part of first-packet */
   last_ts;         /*    and last-packet timestamps */

uint32_t mx_active_streams,
   av_cl_count, n_hc_searches,  src_av_cl_count, src_n_hc_searches;
uint32_t n_not_ip,n_frags,n_too_small,n_no_local_address,
   n_v6_opaque;

uint32_t start_ts;  /* Seconds part of first-packet timestamp */

int snaplen;
uint32_t g_last_ts;  /* Max last_ts for all interfaces */

struct if_info if_info[MXINTERFACES];
uint32_t n_interfaces,   /* Nbr of interfaces specified */
   n_interfaces_running;
char *trace_name, *meter_loc;

owt_queue_declare_space(opq, opq_entry);  /* Define opq itself */

struct event event_q;  /* Queue of active events, dummy first member */

pthread_mutex_t
   log_file_mutex,     /* log_file */
   event_q_mutex,      /* Event queue */
   if_info_mutex,      /* if_info table */
   stream_data_mutex,  /* stream data (packet/byte counts) */
   n_fw_mutex,         /* n_fw_running */
   count_mutex;        /* lock for count/count_cv */

pthread_cond_t count_cv;
int count;  /* Nbr of check_timeouts() running */

int n_fw_running;  /* Nbr of flow watchers running */

Semaphore stream_empty;  /* Controls for packet-watcher input queue */

Semaphore opq_ready, opq_empty;  /* Output queue (to outer-block loop) */

#include <sys/utsname.h>
struct utsname node;  /* Info about the host we're nunning on */

char *version;

int request_shutdown;  /* True to stop stream engine */

int live_source;  /* Zero for trace files */

int my_uid, ut_uid, my_gid, ut_gid;


void quack(int agent) {
   log_msg(LOG_ERR, 0,  "***** quack(%d) *****\n", agent);
   }

void log_msg(int priority, int die, char const *fmt, ...)
{
/* priority = LOG_ERR, LOG_WARNING, LOG_INFO, LOG_DEBUG
      (LOG_INFO messages aren't written to local log file)
      die = +1 to exit, 0 to prefix tod, -1 for no timestamp */

   char msg[300];  /* Longest log message ~ 130, 13 Mar 01 !!! */
   va_list ap;
   time_t t;  char *ts, ftb[32];

   va_start(ap, fmt);
   vsprintf(msg, fmt, ap);
   va_end(ap);

   if (priority != LOG_INFO) {
      mutex_lock("lms", &log_file_mutex);
      setegid(ut_gid);  seteuid(ut_uid);
         if (die == 0) {
            time(&t);  ts = ctime(&t);
            strncpy(&ftb[0],  &ts[11], 9);  /* 17:31:42_ */
            strncpy(&ftb[9],  &ts[0],  4);  /* Thu_ */
            strncpy(&ftb[13], &ts[8],  3);  /* 23_  */
            strncpy(&ftb[16], &ts[4],  4);  /* Sep_ */
            strncpy(&ftb[20], &ts[20], 4);  /* 1997 */
            ftb[24] = '\0';
            fprintf(log_file,"%s (%2d): %s\n", ftb, priority, msg);
            }
         else fprintf(log_file,"%s\n", msg);
      seteuid(my_uid);  setegid(my_uid);
      mutex_unlock("lms", &log_file_mutex);
      }
   fprintf(stderr, "%s\n", msg);
   if (die > 0) {
      stop_meter("### Fatal error, check log file ###");
      }
   }

void uid_from_cwd(void) {
   char buf[200];  int j;
   struct passwd *pw;

   my_uid = getuid();  /* ids we started with (root for a live interface) */
   my_gid = getgid();
   ut_uid = ut_gid = 0;  /* (non-root) ids for stream_engine */
   if (getcwd(buf, sizeof(buf)) != NULL) {  /* Get username from cwd */
      if (strncmp(buf, "/home/", 6) == 0) {
         for (j = 6; buf[j] != '/'; ++j);
         buf[j] = 0;
         if ((pw = getpwnam(&buf[6])) != NULL) {  /* Look up its uid */
            ut_uid = pw->pw_uid;  ut_gid = pw->pw_gid;
	    }
         }
      }
   }

owt_queue_declare_functions(event);
owt_mem_declare_space(event);
owt_mem_declare_functions(event);

struct event *get_event(int t, int interval) {  /* Get an unused event */
   struct event *e = get_event_object(&event_space);
   if (e != NULL) {
      e->type = t;  e->interval = interval;
      return e;
      }
   else return NULL;
   }

void schedule_event(struct event *ev, uint32_t first_t) {
   struct event *eqp, *lqp;

   if (ev == NULL)
      log_msg(LOG_WARNING, 0, "Tried to schedule NULL event!");

   ev->t = first_t;
   mutex_lock("ue", &event_q_mutex);
   lqp = &event_q;  eqp = event_q.next;
   for ( ; eqp != NULL; lqp = eqp, eqp = eqp->next) {
      if (ev->t <= eqp->t) break;  /* Closest to head */
      }
   lqp->next = ev;  ev->next = eqp;
   mutex_unlock("ue", &event_q_mutex);
   }

int check_events(uint32_t now) {
   struct event *ep;
   uint32_t e_type;

   mutex_lock("cde", &event_q_mutex);
   ep = event_q.next;
   if (ep == NULL || ep->t > now) {
      mutex_unlock("cde", &event_q_mutex);
      return NO_EVENT;
      }
   event_q.next = ep->next;  /* Remove from event queue */
   mutex_unlock("cde", &event_q_mutex);
   e_type = ep->type;

   if (ep->interval != 0)
      schedule_event(ep, ep->t += ep->interval);
   else free_event_object(ep, &event_space);

   return e_type;
   }

owt_queue_declare_functions(opq_entry);
owt_mem_declare_space(opq_entry);
owt_mem_declare_functions(opq_entry);

int sp_bad(struct stream *sp) {
   uint32_t ix;
   ix = (uint32_t)(sp - stream_space.objects);
   return ix == 0 || ix >= stream_space.mx_objects;
   }

struct opq_entry *oqe_for_stream(struct stream *sp) {
   struct opq_entry *oqe;
   struct stream *esbp;
   int ix;
   if ((oqe = get_opq_entry_object(&opq_entry_space)) == NULL)
      return NULL;
   if (sp != NULL) {
      esbp = extra_streams[sp->ext_blk];
      ix = (int)(sp - esbp);
      oqe->s_id = sp->ext_blk | (ix << 3);
      }
   else oqe->s_id = -1;
   oqe->next = NULL;
   return oqe;
   }

struct stream *stream_from_oqe(struct opq_entry *oqe) {
   struct stream *esbp = extra_streams[oqe->s_id & 0x7];
   return &esbp[oqe->s_id >> 3];
   }

owt_queue_declare_functions(distribution);
owt_mem_declare_space(distribution);
owt_mem_declare_functions(distribution);

owt_queue_declare_functions(v4_flow_info);
owt_mem_declare_space(v4_flow_info);
owt_mem_declare_functions(v4_flow_info);

owt_queue_declare_functions(v6_flow_info);
owt_mem_declare_space(v6_flow_info);
owt_mem_declare_functions(v6_flow_info);

owt_queue2_declare_functions(stream);
owt_mem_declare_space(stream);
owt_mem_declare_functions(stream);

struct stream *get_stream(int kind)  /* Get an unused stream */
{  /* kind: 4=IPv4 or 6=IPv6 */
   struct stream *sp, *extra_stream_objects;
   int j;
   if (kind == 4 || kind == 6) {
      if (stream_space.free_queue.length != 0)
         sp = get_stream_object(&stream_space);
      else {
 	 if (n_extra_streams == MX_EXTRA_BLOCKS) {
            log_msg(LOG_ERR, 0, "More than %d extra stream blocks needed <<<",
                MX_EXTRA_BLOCKS);
            request_shutdown = 1;
            return NULL;
	    }
         if ((extra_stream_objects = (struct stream *)calloc(
	       extra_stream_incr, sizeof(struct stream))) == NULL)
            log_msg(LOG_ERR, 1, "Couldn't calloc() extra stream objects!");
         n_extra_streams += 1;
         extra_streams[n_extra_streams] = extra_stream_objects;
         stream_space.mx_objects += extra_stream_incr;
         stream_q_lock(&stream_space.free_queue);  /* Lock added, 19 Apr 10 */
         for (j = 0; j != extra_stream_incr; ++j) {
	    extra_stream_objects[j].ext_blk = n_extra_streams;
	    stream_push(&stream_space.free_queue, &extra_stream_objects[j]);
	    }
         stream_q_unlock(&stream_space.free_queue);
	 sp = get_stream_object(&stream_space);
	 log_msg(LOG_WARNING, 0,
            "Extra streams (%d): mx_streams now = %d <<<<<",
	    n_extra_streams, stream_space.mx_objects);
         }
      sp->event_type = 0;  // XXX
      return sp;
      }
   else log_msg(LOG_ERR, 2, "Failed to get_stream(%d) <<<", kind);
   return NULL;  /* Keep com[iler happy */
   }

void get_flow_info(struct stream *sp) {  /* Get flow_info for a stream */
   if (sp->version == 4) {
      if ((sp->flow_info.fi4 = get_v4_flow_info_object(&v4_flow_info_space)
	    ) == NULL) {
	 if (v4_flow_info_space.n_extra_blocks == MX_EXTRA_BLOCKS)
	    return;
 	 make_more_v4_flow_info_space(&v4_flow_info_space);
         if ((sp->flow_info.fi4 = get_v4_flow_info_object(&v4_flow_info_space)
	       ) == NULL)
            return;  /* Don't bother to get distribution */
         }
      init_flow_info(sp);
      if ((sp->flow_info.fi4->d =
            get_distribution_object(&distribution_space)) == NULL) {
	 if (distribution_space.n_extra_blocks == MX_EXTRA_BLOCKS)
	    return;
 	 make_more_distribution_space(&distribution_space);
         if ((sp->flow_info.fi4->d =
	    get_distribution_object(&distribution_space)) == NULL)
	    return;  /* Couldn't get distribution */
	 }
      init_distrib(sp->flow_info.fi4->d);
      }
   else {
      if ((sp->flow_info.fi6 = get_v6_flow_info_object(&v6_flow_info_space)
	    ) == NULL) {
	 if (v6_flow_info_space.n_extra_blocks == MX_EXTRA_BLOCKS)
	    return;
 	 make_more_v6_flow_info_space(&v6_flow_info_space);
         if ((sp->flow_info.fi6 = get_v6_flow_info_object(&v6_flow_info_space)
	       ) == NULL)
            return;  /* Don't bother to get distribution */
         }
      init_flow_info(sp);
      if ((sp->flow_info.fi6->d =
            get_distribution_object(&distribution_space)) == NULL) {
	 if (distribution_space.n_extra_blocks == MX_EXTRA_BLOCKS)
	    return;
 	 make_more_distribution_space(&distribution_space);
         if ((sp->flow_info.fi6->d =
	    get_distribution_object(&distribution_space)) == NULL)
	    return;  /* Couldn't get distribution */
	 }
      init_distrib(sp->flow_info.fi6->d);
      }
   }

void free_flow_info(struct stream *sp) {
   if (sp->version == 4) { 
      if (sp->flow_info.fi4 != NULL) {
 	 if (sp->flow_info.fi4->d != NULL) {
            free_distribution_object(sp->flow_info.fi4->d, &distribution_space);
	    sp->flow_info.fi4->d = NULL;
	    }
 	 free_v4_flow_info_object(sp->flow_info.fi4, &v4_flow_info_space);
         sp->flow_info.fi4 = NULL;
         }
      }
   else if (sp->flow_info.fi6 != NULL) {
      if (sp->flow_info.fi6->d != NULL) {
         free_distribution_object(sp->flow_info.fi6->d, &distribution_space);
	 sp->flow_info.fi6->d = NULL;
         }
      free_v6_flow_info_object(sp->flow_info.fi6, &v6_flow_info_space);
      sp->flow_info.fi6 = NULL;
      }
   }

void free_stream(struct stream *sp)  /* Return stream to free list */
{
   sp->state = SS_FREE;
   free_flow_info(sp);
   free_stream_object(sp, &stream_space);
   }

void stream_fl_check(void) {
   int j;
   struct stream *csp;

   printf("fl_check(): n_free_stream=%d, stream_space=%p",
      stream_space.free_queue.length, stream_space.objects);
   for (j = 1, csp = stream_space.free_queue.head; csp->next != NULL; ) {
      j += 1;  csp = csp->next;
      }
   if (j == stream_space.free_queue.length) printf(": OK\n");
   else printf(": error, %d != %d\n", j, stream_space.free_queue.length);

   fflush(stdout);
   }

owt_queue_declare_functions(stream_txt)
owt_mem_declare_space(stream_txt)
owt_mem_declare_functions(stream_txt)

owt_queue_declare_functions(flow_watcher_control)
owt_mem_declare_space(flow_watcher_control)
owt_mem_declare_functions(flow_watcher_control)

void init_mutex(pthread_mutex_t *m, char const *name) {
   if (pthread_mutex_init(m, pthread_mutexattr_default) == -1)
      log_msg(LOG_ERR, 6, "Couldn't init %s", name);
   }

void init_mutexes(void) {
   init_mutex(&log_file_mutex, "log_file_mutex");
   init_mutex(&event_q_mutex, "event_q_mutex");
   init_mutex(&if_info_mutex, "if_info_mutex");
   init_mutex(&stream_data_mutex, "stream_data_mutex");
   init_mutex(&n_fw_mutex, "n_fw_mutex");
   init_mutex(&count_mutex, "count_mutex");
   }

void owt_start(int mx_events, int mx_opq_entries, int mx_streams,
   int mx_stream_txts, int mx_flow_watcher_controls)
{
   n_unanalysed_sources = 0;  /* Un-analysed source count */
   init_mutexes();
   init_event_space(&event_space, mx_events);
   memset(&event_q, 0, sizeof(struct event));
   event_q.type = NO_EVENT;

   init_opq_entry_space(&opq_entry_space, mx_opq_entries);
   init_opq_entry_queue(&opq);

   init_stream_space(&stream_space, mx_streams);
   extra_streams[0] = stream_space.objects;
   extra_stream_incr = mx_streams;  
   n_extra_streams = 0;

   init_v4_flow_info_space(&v4_flow_info_space, mx_streams);
   init_v6_flow_info_space(&v6_flow_info_space, mx_streams/8);
 
   init_distribution_space(&distribution_space,
      v4_flow_info_space.mx_objects + v6_flow_info_space.mx_objects);

   init_stream_txt_space(&stream_txt_space, mx_stream_txts);

   init_flow_watcher_control_space(
      &flow_watcher_control_space, mx_flow_watcher_controls);

#if HALF_WAY_STREAMS
   half_src_addrs = NULL;
#endif
   }

void owt_stop(void) {
   }


static char *strmov(char *d, char *s) {
   while (*s != '\0') *d++ = *s++;
   return d;
   }

char *v6addr_to_s(char *v6a, struct ip_address *p)
{  /* Returns pointer to next byte in v6a 
      Code from NeTraMet's nmc_pars.c */
   char buf[10];  /* RFC 2373: IPv6 Address Architecture */
   char *d = v6a;
   char *a = (char *)p->a.a;
   int j, k, st,len, stx = 0,lenx;
   uint32_t v, a2[8];

   st = len = lenx = 0;
   for (k = j = 0; j != 16; j += 2) {
      v =  ntohs(*(uint16_t *)&a[j]);
      a2[k++] = v;          /* Build array of two-byte pairs */
      if (v == 0) ++len;
      else {
         if (len > lenx) {  /* Find longest run of zero pairs */
            stx = st;  lenx = len;
	    }
         st = k;  len = 0;
         }
      }

   if (len > lenx) {
      stx = st;  lenx = len;
      }
   if (lenx != 0 && stx == 0) {  /* Longest run at left */
     d = strmov(d, (char *)":");  j = lenx;
      }
   else {
      sprintf(buf, "%x", a2[0]);
      d = strmov(d,buf);  j = 1;
      }
   for (; j < 8; ) {
      if (lenx != 0 && j == stx) {
	d = strmov(d,(char *)":");  j += lenx;
         }
      else {
         sprintf(buf, ":%x", a2[j]);
         d = strmov(d, buf);  ++j;
         }
      }
   if (j == stx+lenx) d = strmov(d, (char *)":");  /* Longest run at right */
   *d = '\0';
   return d;
   }


struct owt_ht *owt_ht_create(int mx_hash_entries) {
   struct owt_ht *new_htbl = malloc(sizeof(struct owt_ht));
   init_mutex(&new_htbl->ht_lock, "owt_ht_create");
   new_htbl->table_size = mx_hash_entries;
   new_htbl->sfht = (void **)malloc(mx_hash_entries*sizeof(void *));
   if (new_htbl->sfht == NULL)
      log_msg(LOG_ERR, 7, "Not enough memory for stream hash table!");
   memset(new_htbl->sfht, 0, new_htbl->table_size*sizeof(void *));
   return new_htbl;
   }

uint32_t owt_hash(struct stream *sp) {
    /* stream hash, determines flow_watcher for every packet,
          also determines stream the packet belongs to */
   if (sp->version == 4) {
     //      return sp->src_addr.a.v4 ^ (4 * 97171);
      return (sp->src_addr.a.v4 >> 9) + sp->src_addr.a.v4 * 5;
      }
   else {  /* v6 stream */
      uint64_t hh = 
 	 sp->src_addr.a.o.oct0 ^ sp->src_addr.a.o.oct1;
      return ((uint32_t *)&hh)[1] ^ ((uint32_t *)&hh)[0] ^ (6 * 96953);
      }
   }

uint32_t owt_compare(void *record, void *sh_entry) {
   struct stream *sp = (struct stream *)record;
   struct stream *shte = (struct stream *)sh_entry;
   if (sp->version == 4) {
      return
	 sp->src_addr.a.v4 == shte->src_addr.a.v4 &&
         sp->version == shte->version;
      }
   else {  /* v6 stream */
      return 
         sp->src_addr.a.o.oct1 == shte->src_addr.a.o.oct1 &&
	 sp->src_addr.a.o.oct0 == shte->src_addr.a.o.oct0 &&
         sp->version == shte->version;
      }
   }

void *owt_ht_lookup(struct owt_ht *t, uint32_t hash,
      struct stream *record, int *OK_to_add) {
   struct stream **sha;
   struct stream *shte, *lshte = NULL, *sp;
   int cl;

   mutex_lock("slu", &t->ht_lock);
   sha = (struct stream **)&t->sfht[hash % t->table_size];
      /* Address of stream hash table entry */
   if ((shte = *sha) == NULL) {
      lshte = NULL;  /* Empty hash chain */
      }
   else {
      for (cl = 1; ; cl += 1) {
 	 if (owt_compare(record, shte)) break;  /* Matched */
         lshte = shte;
         if ((shte = lshte->next_hc) == (struct stream *)sha) {
            shte = NULL;  break;
            /* Loop stops with lshte -> last stream in chain */
	    }
         }
      av_cl_count += cl;  n_hc_searches += 1;
      }

   if (shte != NULL) {  /* Existing stream */
      *OK_to_add = 0;
//    if (shte == record) quack(6001);
      mutex_unlock("slu", &t->ht_lock);
      return shte;
      }
   else if (!*OK_to_add) {  /* Not in stream table */
      mutex_unlock("slu", &t->ht_lock);
      return NULL;
      }
   else {  /* Enter record in stream table */
      sp = (struct stream *)record;
//    if (shte != NULL) quack(6002);
      if (lshte == NULL) *sha = sp;  /* Add stream to hash chain */
      else lshte->next_hc = sp;
      sp->next_hc = (struct stream *)sha;
      sp->state = SS_ONE_WAY;
      //?? sp->last_report_sec = sp->ts;   //??
      mutex_unlock("slu", &t->ht_lock);
      *OK_to_add = 1;  /* We added this stream */
      return sp;
      }
   }

void owt_ht_delete(struct owt_ht *t, void *record) {
   struct stream *sp, *np, *tp, *ltp;

   sp = (struct stream *)record;
   mutex_lock("slu", &t->ht_lock);
   np = sp->next_hc;  /* Remove stream from hash chain */
   tp = sp;  do {
      ltp = tp;
      } while ((tp = tp->next_hc) != sp);
   if (ltp == np)  /* Single flow in hash chain */
      np->next_hc = NULL;
   else ltp->next_hc = np;
   sp->next_hc = NULL;
   mutex_unlock("slu", &t->ht_lock);
   }

static uint8_t p_mask[8] = { 0x00, 0x80, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0xFE };

int addr_is_local(int ver, struct ip_address *addrp) {  /* in local_addrs[] */
   int j, nb,k, len;
   uint8_t *a = addrp->a.a;
   for (j = 0; j != n_local_addrs; ++j) {
      if (ver == la[j].ver) {
	 len = la[j].len;
         nb = len/8;  /* whole bytes */
         for (k = 0; k != nb; ++k)
            if (a[k] != la[j].a.a[k]) break;
         if (k != nb) return 0;  /* false */
         if (k*8 == len) return 1;  /* Match up to length */
	 // if (addrp->ver == 6) { 
	 //    printf("   =========== last byte match %d  %02x ^ %02x -> %d\n",
	 //    k, a[k], la[j].a.a[k], ((a[k] ^ la[j].a.a[k]) & p_mask[len%8]) == 0);
	 //    }
         return ((a[k] ^ la[j].a.a[k]) & p_mask[len%8]) == 0;
         }
      }   
   return 0;
   }

void reset_summary_info(void) {
   uint32_t j;
   v4_one_way_bytes     = v4_one_way_pkts  = v4_one_way_streams =
      v6_one_way_bytes  = v6_one_way_pkts  = v6_one_way_streams = 
      v4_two_way_bytes  = v4_two_way_pkts  = v4_two_way_streams =
      v6_two_way_bytes  = v6_two_way_pkts  = v6_two_way_streams =
      v4_half_way_bytes = v4_half_way_pkts = v4_half_way_streams =
      v6_half_way_bytes = v6_half_way_pkts = v6_half_way_streams = 0;
#if ERF_DIRECTION
   one_way_streams_1 = one_way_streams_2 = one_way_streams_3 =
      two_way_streams_1 = two_way_streams_2 = two_way_streams_3 =
      half_way_streams_1 = half_way_streams_2 = half_way_streams_3 = 0;
#endif
#if HALF_WAY_STREAMS
   half_src_addrs = create_hashtable(5000, hashfromaddr, equaladdrs);
#endif
#if MKDESTCOUNTS
   for (j = 0; j != n_source_types; ++j) {
      memset(dest_addrs_per_source_type[j], 0, NDESTBINS*sizeof(uint32_t));
      }
#endif
   }

void update_summary_info(struct stream *sp) {
#if ERF_DIRECTION
   update_direction_counts(sp);
#endif
   if (sp->state == SS_TWO_WAY) {
      if (sp->version == 4) {
	 v4_two_way_bytes += (sp->tb + sp->fb);
	 v4_two_way_pkts += (sp->tp + sp->fp);
         v4_two_way_streams += 1;
 	 }
      else {
         v6_two_way_bytes += (sp->tb + sp->fb);
         v6_two_way_pkts += (sp->tp + sp->fp);
         v6_two_way_streams += 1;
         }
      }
#if HALF_WAY_STREAMS
   else if (sp->state == SS_HALF_WAY) {
      if (sp->version == 4) {
	 v4_half_way_bytes += (sp->tb + sp->fb);
	 v4_half_way_pkts += (sp->tp + sp->fp);
         v4_half_way_streams += 1;
         addr_ht_insert(half_src_addrs, &sp->src_addr);
 	 }
      else {
         v6_half_way_bytes += (sp->tb + sp->fb);
         v6_half_way_pkts += (sp->tp + sp->fp);
         v6_half_way_streams += 1;
         addr_ht_insert(half_src_addrs, &sp->src_addr);
         }
      }
#endif
    }


void init_flow_info(struct stream *sp) {  /* Zero flow_info for a stream */
   if (sp->version == 4) {
      struct v4_flow_info *fi4 = sp->flow_info.fi4;
      fi4->cfc_count = fi4->tn_tcp.nv = fi4->tn_udp.nv =
         fi4->tn_flags.nv = fi4->tn_proto.nv = fi4->tn_ttl.nv =
	 fi4->tn_dst_addrs.nv = fi4->tn_src_port.nv = 0;
#if HALF_WAY_STREAMS
      fi4->pkts_this_dst = 0;
#endif
      }
   else {
      struct v6_flow_info *fi6 = sp->flow_info.fi6;
      fi6->cfc_count = fi6->tn_tcp.nv = fi6->tn_udp.nv =
         fi6->tn_flags.nv = fi6->tn_proto.nv = fi6->tn_ttl.nv =
 	 fi6->tn_dst_addrs.nv = fi6->tn_src_port.nv = 0;
#if HALF_WAY_STREAMS
      fi6->pkts_this_dst = 0;
#endif
      }
   }

void count_uint16(uint16_t p, struct topn_uint16 *tnp) {
   uint32_t j;
   for (j = 0; j != tnp->nv; j += 1) {
      if (p == tnp->values[j]) {
	 tnp->counts[j] += 1;  return;
	 }
      }
   if (j != MX_N_UINT16) {
      tnp->values[j] = p;  tnp->counts[j] = 1;
      tnp->nv += 1;
      }
   }

void count_uint32(uint32_t p, struct topn_uint32 *tnp) {
   uint32_t j;
   for (j = 0; j != tnp->nv; j += 1) {
      if (p == tnp->values[j]) {
	 tnp->counts[j] += 1;  return;
	 }
      }
   if (j != MX_N_UINT16) {
      tnp->values[j] = p;  tnp->counts[j] = 1;
      tnp->nv += 1;
      }
   }

void count_uint8(uint8_t p, struct topn_uint8 *tnp) {
   uint32_t j;
   for (j = 0; j != tnp->nv; j += 1) {
      if (p == tnp->values[j]) {
	 tnp->counts[j] += 1;  return;
	 }
      }
   if (j != MX_N_UINT8) {
      tnp->values[j] = p;  tnp->counts[j] = 1;
      tnp->nv += 1;
      }
   }

void count_addrs4(struct ip_address *p, struct topn_dst_addrs4 *tnp) {
   uint32_t j;
   for (j = 0; j != tnp->nv; j += 1) {
      if (p->a.v4 == tnp->values[j].a.v4) {
	 tnp->counts[j] += 1;  return;
	 }
      }
   if (j != V4_SET_SIZE) {
      tnp->values[j].a.v4 = p->a.v4;  tnp->counts[j] = 1;
      tnp->nv += 1;
      }
   }

void count_addrs6(struct ip_address *p, struct topn_dst_addrs6 *tnp) {
   uint32_t j;
   for (j = 0; j != tnp->nv; j += 1) {
      if (p->a.o.oct0 == tnp->values[j].a.o.oct0 &&
            p->a.o.oct1 == tnp->values[j].a.o.oct1) {
	 tnp->counts[j] += 1;  return;
	 }
      }
   if (j != V6_SET_SIZE) {
      tnp->values[j].a.o.oct0 = p->a.o.oct0;
      tnp->values[j].a.o.oct1 = p->a.o.oct1;
      tnp->counts[j] = 1;  tnp->nv += 1;
      }
   }

uint8_t top_uint8(struct topn_uint8 *tnp) {
   if (tnp->nv == 0) return 0;
   uint16_t j, mj = 0;
   uint32_t mxcount = tnp->counts[0];
   for (j = 1; j != tnp->nv; j += 1) {
      if (tnp->counts[j] > mxcount) {
         mj = j;  mxcount = tnp->counts[j];
	 }
      }
   return tnp->values[mj];
   }

uint16_t top_uint16(struct topn_uint16 *tnp) {
   if (tnp->nv == 0) return 0;
   uint32_t j, mj = 0,
      mxcount = tnp->counts[0];
   for (j = 1; j != tnp->nv; j += 1) {
     if (tnp->counts[j] > mxcount) {
         mj = j;  mxcount = tnp->counts[j];
	 }
      }
   return tnp->values[mj];
   }

uint32_t top_uint32(struct topn_uint32 *tnp) {
   if (tnp->nv == 0) return 0;
   uint32_t j, mj = 0,
     mxcount = tnp->counts[0];
   for (j = 1; j != tnp->nv; j += 1) {
     if (tnp->counts[j] > mxcount) {
         mj = j;  mxcount = tnp->counts[j];
	 }
      }
   return tnp->values[mj];
   }

uint32_t top_addrs4(struct topn_dst_addrs4 *tnp) {
   if (tnp->nv == 0) return 0;
   uint32_t j, mj = 0,
     mxcount = tnp->counts[0];
   for (j = 0; j != tnp->nv; j += 1) {
     if (tnp->counts[j] > mxcount) {
         mj = j;  mxcount = tnp->counts[j];
	 }
      }
   return tnp->values[mj].a.v4;
   }

int top2_uint8(struct topn_uint8 *tnp, int *v2) {
   if (tnp->nv == 0) {
      *v2 = -1;  return -1;
      }
   uint32_t j, mj = 0, mj2 = 0,
      mxcount = tnp->counts[0],
      mx2 = 0;
   if (tnp->nv == 0) {
      *v2 = -1; return -1;
      }
   for (j = 1; j != tnp->nv; j += 1) {
      if (tnp->counts[j] > mxcount) {
         mj2 = mj;  mx2 = mxcount;
         mj = j;  mxcount = tnp->counts[j];
	 }
      else if (tnp->counts[j] > mx2) {
         mj2 = j;  mx2 = tnp->counts[j];
         }
      }
   *v2 = mx2 == 0 ? -1 : tnp->values[mj2];
   return tnp->values[mj];
   }

int top2_uint16(struct topn_uint16 *tnp, int *v2) {
   if (tnp->nv == 0) {
      *v2 = -1;  return -1;
      }
   uint32_t j, mj = 0, mj2 = 0,
      mxcount = tnp->counts[0],
      mx2 = 0;
   if (tnp->nv == 0) {
      *v2 = -1; return -1;
      }
   for (j = 1; j != tnp->nv; j += 1) {
      if (tnp->counts[j] > mxcount) {
         mj2 = mj;  mx2 = mxcount;
         mj = j;  mxcount = tnp->counts[j];
	 }
      else if (tnp->counts[j] > mx2) {
         mj2 = j;  mx2 = tnp->counts[j];
         }
      }
   *v2 = mx2 == 0 ? -1 : tnp->values[mj2];
   return tnp->values[mj];
   }

int update_flow_info(struct stream *sp, struct stream *isp)
{  /* Update sp flow info from 'this packet' fields in isp,
      returns 1 if we've seen increasing TCP seq or ack */
   if (sp->version == 4) {
      struct v4_flow_info *fi4 = sp->flow_info.fi4;
      if (fi4->d != NULL) bump_dist(fi4->d, isp->last_us);
      count_uint8(isp->proto, &fi4->tn_proto);
      fi4->cfc_count += isp->cfc;
      fi4->uT_count += isp->uTorrent;
      switch (isp->proto) {
      case PT_TCP:
         count_uint16(isp->dst_port, &fi4->tn_tcp);
         count_uint16(isp->src_port, &fi4->tn_src_port);
         count_uint8(isp->tcp_flags, &fi4->tn_flags);
#if HALF_WAY_STREAMS
         if (fi4->pkts_this_dst == 0) {  /* New source */
 	    fi4->prev_dst.a.v4 = isp->dst_addr.a.v4;
            fi4->prev_seq = isp->tcp_seq;
            fi4->prev_ack = isp->tcp_flags & 0x10 ? isp->tcp_ack :
               0;  /* No ACK flag, guess first ACK value (!) */
            fi4->pkts_this_dst = 1;
	    }
         else {
	    if (isp->dst_addr.a.v4 != fi4->prev_dst.a.v4) {
 	       fi4->prev_dst.a.v4 = isp->dst_addr.a.v4;  /* New dst */
               fi4->prev_seq = isp->tcp_seq;
               fi4->prev_ack = isp->tcp_ack;
               fi4->pkts_this_dst = 1;
	       }
 	    else if (isp->tcp_seq > fi4->prev_seq) {
	       fi4->prev_seq = isp->tcp_seq;  fi4->pkts_this_dst += 1;
	       }
            else if ((isp->tcp_flags & 0x10) && isp->tcp_ack > fi4->prev_ack) {
	       fi4->prev_ack = isp->tcp_ack;  fi4->pkts_this_dst += 1;
	       }
            }
#endif
         break;
      case PT_UDP:
         count_uint16(isp->dst_port, &fi4->tn_udp);
         count_uint16(isp->src_port, &fi4->tn_src_port);
         break;
      case PT_ICMP:
         count_uint16(isp->dst_port, &fi4->tn_tcp);  /* ICMP type|code */
         break;
         }
      count_uint8(isp->ttl, &fi4->tn_ttl);
      count_addrs4(&isp->dst_addr, &fi4->tn_dst_addrs);
#if HALF_WAY_STREAMS
      return fi4->pkts_this_dst >= HALF_ONE_WAY_PKTS;
#else
      return 0;
#endif
      }
   else {
      struct v6_flow_info *fi6 = sp->flow_info.fi6;
      if (fi6->d != NULL) bump_dist(fi6->d, isp->last_us);
      fi6->cfc_count += isp->cfc;
      fi6->uT_count += isp->uTorrent;
      count_uint8(isp->proto, &fi6->tn_proto);
      switch (isp->proto) {
      case PT_TCP:  
         count_uint16(isp->dst_port, &fi6->tn_tcp);
         count_uint16(isp->src_port, &fi6->tn_src_port);
         count_uint8(isp->tcp_flags, &fi6->tn_flags);
#if HALF_WAY_STREAMS
         if (fi6->pkts_this_dst == 0) {  /* New source */
  	    fi6->prev_dst.a.o.oct0 = isp->dst_addr.a.o.oct0;
  	    fi6->prev_dst.a.o.oct1 = isp->dst_addr.a.o.oct1;
            fi6->prev_seq = isp->tcp_seq;  fi6->prev_ack = isp->tcp_ack;
            fi6->pkts_this_dst = 1;
	    }
         else {
	    if (isp->dst_addr.a.o.oct0 != fi6->prev_dst.a.o.oct0 ||
                  isp->dst_addr.a.o.oct1 != fi6->prev_dst.a.o.oct1) {
  	       fi6->prev_dst.a.o.oct0 = isp->dst_addr.a.o.oct0;
  	       fi6->prev_dst.a.o.oct1 = isp->dst_addr.a.o.oct1;
               fi6->prev_seq = isp->tcp_seq;
               fi6->prev_ack = isp->tcp_ack;
               fi6->pkts_this_dst = 1;
	       }
 	    else if (isp->tcp_seq > fi6->prev_seq) {
	       fi6->prev_seq = isp->tcp_seq;  fi6->pkts_this_dst += 1;
	       }
            else if (isp->tcp_ack > fi6->prev_ack) {
	       fi6->prev_ack = isp->tcp_ack;  fi6->pkts_this_dst += 1;
	       }
            }
#endif
         break;
      case PT_UDP:
         count_uint16(isp->dst_port, &fi6->tn_udp);
         count_uint16(isp->src_port, &fi6->tn_src_port);
         break;
      case PT_6ICMP:
         count_uint16(isp->dst_port, &fi6->tn_tcp);  /* ICMP type|code */
         break;
         }
      count_uint8(isp->ttl, &fi6->tn_ttl);
      count_addrs6(&isp->dst_addr, &fi6->tn_dst_addrs);
#if HALF_WAY_STREAMS
      return fi6->pkts_this_dst >= HALF_ONE_WAY_PKTS;
#else
      return 0;
#endif
      }
   }

#define S_unclassified          0

#define S_tcp_probe             1
#define S_tcp_vertical_scan     2
#define S_tcp_horizontal_scan   3
#define S_tcp_unknown           4

#define S_udp_probe             5
#define S_udp_vertical_scan     6
#define S_udp_horizontal_scan   7
#define S_udp_unknown           8

#define S_icmp_only             9
#define S_backscatter          10
#define S_tcp_and_udp          11

#define S_utorrent             12
#define S_conficker_c          13
#define S_ipsec                14

#define CFC_FRACTION          3/4


#define IAT_Other               0  /* group numbers from classify() in t9.rb */
#define IAT_short_lived         1
#define IAT_high_rate           2  /* > 5 pkt/s */
#define IAT_DOS                 3

#define IAT_3_left              4
#define IAT_3_even              5
#define IAT_3_right             6

#define IAT_stealth_3s_mode     7  /* Long-lived, few pkts */
#define IAT_stealth_spikes      8
#define IAT_stealth             9


uint16_t probe_port, src_port;  /* analyse_flow() results */
int flags_mode, keys_mode;
uint8_t probe_port_valid,   /* (other than src_type) */
   too_few_pkts,      /* < MinPkts */
   protocol;

double b0pc, modeiat, skew;  /* compute_distrib_stats() results */
int nz_runs;

int analyse_flow(struct stream *sp)
{  /* Determine source kind from flow info */
   too_few_pkts = sp->totp < min_pkts_to_classify;
   probe_port_valid = 0;
   flags_mode = keys_mode = src_port = 0;
   if (sp->version == 4) {
      struct v4_flow_info *fi4 = sp->flow_info.fi4;
      if (fi4->tn_proto.nv == 1) {  /* Single protocol */
	 protocol = fi4->tn_proto.values[0];
	 if (protocol == PT_TCP) {
            flags_mode = top_uint8(&fi4->tn_flags);
            keys_mode = sp->tcp_key;
            if (flags_mode == 18 || flags_mode == 20)
               /* TCP ACK+SYN or ACK+RST */
               return S_backscatter;
 	    if (!too_few_pkts) {
	       if (fi4->cfc_count >= sp->totp*CFC_FRACTION) return S_conficker_c;
	       if (fi4->uT_count >= sp->totp*CFC_FRACTION) return S_utorrent;
	       }
	    if (fi4->tn_tcp.nv == 1) {  /* Single dst_port */
	       probe_port = top_uint16(&fi4->tn_tcp);
               probe_port_valid = 1;
               src_port = top_uint16(&fi4->tn_src_port);
               if (src_port == PN_DNS) return S_backscatter;
	       if (fi4->tn_dst_addrs.nv == 1) return S_tcp_probe;
	       return S_tcp_horizontal_scan;
	       }
	    else {  /* >1 dst_port */
               if (fi4->tn_dst_addrs.nv == 1) return S_tcp_vertical_scan;
	       return S_tcp_unknown;
	       }
	    }
	 if (protocol == PT_UDP) {
 	    if (!too_few_pkts) {
	       if (fi4->cfc_count >= sp->totp*CFC_FRACTION) return S_conficker_c;
	       if (fi4->uT_count >= sp->totp*CFC_FRACTION) return S_utorrent;
	       }
	    if (fi4->tn_udp.nv == 1) {  /* Single dst_port */
	       probe_port = top_uint16(&fi4->tn_udp);
               probe_port_valid = 1;
               src_port = top_uint16(&fi4->tn_src_port);
               if (src_port == PN_DNS) return S_backscatter;
               if (fi4->tn_dst_addrs.nv == 1) return S_udp_probe;
	       return S_udp_horizontal_scan;
	       }
	    else {  /* >1 dst_port */
               if (fi4->tn_dst_addrs.nv == 1) return S_tcp_vertical_scan;
	       return S_udp_unknown;
	       }
            }
	 if (protocol == PT_ICMP) {
            int it = top_uint16(&fi4->tn_tcp);  /* ICMP type */
            if (it == ICMP_TTL_EXCEEDED || it == ICMP_DST_UNREACHABLE)
               return S_backscatter;
            return S_icmp_only;
	    }
         if (protocol == PT_6ESP) return S_ipsec;
         return S_unclassified;
         }
      else {
         protocol = top_uint8(&fi4->tn_proto);
         if (fi4->tn_proto.nv == 2 && (
               (fi4->tn_proto.values[0] == PT_TCP &&
                  fi4->tn_proto.values[1] == PT_UDP) ||
	       (fi4->tn_proto.values[0] == PT_UDP &&
	          fi4->tn_proto.values[1] == PT_TCP) ) ) {
 	    if (!too_few_pkts) {
	       if (fi4->cfc_count >= sp->totp*CFC_FRACTION) return S_conficker_c;
	       if (fi4->uT_count >= sp->totp*CFC_FRACTION) return S_utorrent;
	       }
            else return S_tcp_and_udp;
            }
         switch (protocol) {
         case PT_TCP:  
            return S_tcp_unknown;
         case PT_UDP:
            return S_udp_unknown;
         default:
	    return S_unclassified;
            }
         }
      }
   else {
      struct v6_flow_info *fi6 = sp->flow_info.fi6;
      if (fi6->tn_proto.nv == 1) {  /* Single protocol */
	 protocol = fi6->tn_proto.values[0];
	 if (protocol == PT_TCP) {
            flags_mode = top_uint8(&fi6->tn_flags);
            keys_mode = sp->tcp_key;
            if (flags_mode == 18 || flags_mode == 20)
               /* TCP ACK+SYN or ACK+RST */
               return S_backscatter;
 	    if (!too_few_pkts) {
	       if (fi6->cfc_count >= sp->totp*CFC_FRACTION) return S_conficker_c;
	       if (fi6->uT_count >= sp->totp*CFC_FRACTION) return S_utorrent;
	       }
	    if (fi6->tn_tcp.nv == 1) {  /* Single dst_port */
	       probe_port = top_uint16(&fi6->tn_tcp);
               probe_port_valid = 1;
               src_port = top_uint16(&fi6->tn_src_port);
               if (src_port == PN_DNS) return S_backscatter;
	       if (fi6->tn_dst_addrs.nv == 1) return S_tcp_probe;
	       return S_tcp_horizontal_scan;
	       }
	    else {  /* >1 dst_port */
               if (fi6->tn_dst_addrs.nv == 1) return S_tcp_vertical_scan;
	       return S_tcp_unknown;
	       }
	    }
	 if (protocol == PT_UDP) {
 	    if (!too_few_pkts) {
	       if (fi6->cfc_count >= sp->totp*CFC_FRACTION) return S_conficker_c;
	       if (fi6->uT_count >= sp->totp*CFC_FRACTION) return S_utorrent;
	       }
	    if (fi6->tn_udp.nv == 1) {  /* Single dst_port */
	       probe_port = top_uint16(&fi6->tn_udp);
               probe_port_valid = 1;
               src_port = top_uint16(&fi6->tn_src_port);
               if (src_port == PN_DNS) return S_backscatter;
               if (fi6->tn_dst_addrs.nv == 1) return S_udp_probe;
	       return S_udp_horizontal_scan;
	       }
	    else {  /* >1 dst_port */
               if (fi6->tn_dst_addrs.nv == 1) return S_tcp_vertical_scan;
	       return S_udp_unknown;
	       }
            }
	 if (protocol == PT_ICMP) {
            int it = top_uint16(&fi6->tn_tcp);  /* ICMP type */
            if (it == ICMP_TTL_EXCEEDED || it == ICMP_DST_UNREACHABLE)
               return S_backscatter;
            return S_icmp_only;
	    }
         if (protocol == PT_6ESP) return S_ipsec;
         return S_unclassified;
         }
      else {
	 protocol = top_uint8(&fi6->tn_proto);
         if (fi6->tn_proto.nv == 2 && (
               (fi6->tn_proto.values[0] == PT_TCP &&
                  fi6->tn_proto.values[1] == PT_UDP) ||
	       (fi6->tn_proto.values[0] == PT_UDP &&
	          fi6->tn_proto.values[1] == PT_TCP) ) ) {
 	    if (!too_few_pkts) {
	       if (fi6->cfc_count >= sp->totp*CFC_FRACTION) return S_conficker_c;
	       if (fi6->uT_count >= sp->totp*CFC_FRACTION) return S_utorrent;
	       }
            else return S_tcp_and_udp;
            }
         switch (protocol) {
         case PT_TCP:  
            return S_tcp_unknown;
         case PT_UDP:
            return S_udp_unknown;
         default:
	    return S_unclassified;
            }
         }
      }
   return S_unclassified;
   }


char key_s[150];  /* (Global) return value for key_to_s() */
char *key_to_s(struct stream *sp) {
   if (sp->version == 4) {
      uint8_t *a = sp->src_addr.a.a;
      sprintf(key_s, "%u.%u.%u.%u", a[0],a[1],a[2],a[3]);
      }
   else if (sp->version == 6) {
      char *kp = v6addr_to_s(key_s, &sp->src_addr);
      kp[1] = '\0';
      }
   else {
      quack(999);
      log_msg(LOG_ERR, 2,
          "Tried to print key_type %d <<<", sp->version);
      }
   return key_s;
   }

void print_topn_uint8(struct topn_uint8 *tnp) {
   uint32_t j;
   printf("<n=%d,", tnp->nv);
   for (j = 0; j != tnp->nv;  j += 1)
      printf(" %u,%u", tnp->values[j], tnp->counts[j]);
   printf(">");
   }

void print_topn_uint16(struct topn_uint16 *tnp) {
   uint32_t j;
   printf("<n=%d,", tnp->nv);
   for (j = 0; j != tnp->nv;  j += 1)
      printf(" %u,%u", tnp->values[j], tnp->counts[j]);
   printf(">");
   }

void print_topn_dst_addrs4(struct topn_dst_addrs4 *tnp) {
   uint32_t j;
   printf("[n=%d,", tnp->nv);
   for (j = 0; j != tnp->nv;  j += 1) {
      uint8_t *a = tnp->values[j].a.a;
      printf(" %u.%u.%u.%u,%u", a[0],a[1],a[2],a[3], tnp->counts[j]);
      }
   printf("]");
   }

void print_topn_dst_addrs6(struct topn_dst_addrs6 *tnp) {
   uint32_t j;
   printf("[n=%d,", tnp->nv);
   for (j = 0; j != tnp->nv;  j += 1) {
      char *kp = v6addr_to_s(key_s, (struct ip_address *)&tnp->values[j]);
      kp[1] = '\0';
      printf(" %s,%u", key_s, tnp->counts[j]);
      }
   printf("]");
   }

void print_flow_info(struct stream *sp) {
   char *src_addr = key_to_s(sp); 
   printf("%s  P", src_addr);
   if (sp->version == 4) {
      struct v4_flow_info *fi4 = sp->flow_info.fi4;
      print_topn_uint8(&fi4->tn_proto);  printf("  T");
      print_topn_uint16(&fi4->tn_tcp);  printf("  U");
      print_topn_uint16(&fi4->tn_udp);  printf("  A");
      print_topn_dst_addrs4(&fi4->tn_dst_addrs);
      }
   else {
      struct v6_flow_info *fi6 = sp->flow_info.fi6;
      print_topn_uint8(&fi6->tn_proto);  printf("  T");
      print_topn_uint16(&fi6->tn_tcp);  printf("  U");
      print_topn_uint16(&fi6->tn_udp);  printf("  A");
      print_topn_dst_addrs6(&fi6->tn_dst_addrs);
      }
   }


#include <math.h>

void init_distrib(struct distribution *d)
{
   double N, D;

   memset(d, 0, sizeof(struct distribution));

#if 0 //  BURST_STATS
   d->Transform = DS_LOG;
   d->LowerLim = (uint16_t)BURST_GAP;
   d->UpperLim = 3000;
#endif
   d->Transform = DS_LOG;
   d->LowerLim = LOWER_IAT;  d->UpperLim = UPPER_IAT;  /* seconds */
   d->Buckets = MXBUCKETS;  /* Use them all! */

   if (d->Buckets > MXBUCKETS) {
      log_msg(LOG_ERR, 0,
         "Distribution with > %d buckets **", MXBUCKETS);
      d->Buckets = MXBUCKETS;
      }

   N = d->Buckets - 1;
   switch (d->Transform) {
   case DS_LIN:
      D = d->UpperLim - d->LowerLim;
      d->M = N/D;
      d->step = (d->UpperLim - d->LowerLim)/(d->Buckets-1);
      break;
   case DS_LOG:
      D = log10(d->UpperLim/d->LowerLim);
      d->M = N/D;
      d->step = (log10(d->UpperLim) - log10(d->LowerLim))/(d->Buckets-1);
      break;
#if 0
   case DS_DYN_REQ:
   case DS_DLOG_REQ:
      d->dyn_dist = 1;
      d->max_val = d->LowerLim;
      d->min_val = d->UpperLim;
      d->n_values = 0;
      break;
#endif
      }
   }

void scale_dist_limits(struct distribution *d,
   unsigned int ilow, unsigned int ihigh)
{
#if 0  // <<<<<<<<<<<
   uint32_t LowLim, HiLim;
   LowLim = ilow < d->lowerspec ? d->lowerspec : ilow;
   if (LowLim > (uint16_t)~0) LowLim = d->lowerspec;
   d->LowerLim = LowLim;

   d->UpperLim = ihigh > d->upperspec ? d->upperspec : ihigh;
   if (d->HiDel != 0) {
      HiLim =  /* met_vars.c exports UpperLimit */
         ceil(d->UpperLim*pwr_of_10[d->HiDel > 9 ? 9 : d->HiDel]);
      d->UpperLim = times_pwr_of_10(HiLim, d->HiDel);
      }
   else HiLim = d->UpperLim;
   if (HiLim > (uint16_t)~0) {
      d->UpperLimit = d->upperspec*pwr_of_10[d->HiDel > 9 ? 9 : d->HiDel];
      d->UpperLim = times_pwr_of_10(d->UpperLimit, d->HiDel);
      }
   else d->UpperLimit = d->UpperLim;
#endif
   }

#if 0 // BURST_STATS
void end_burst(struct distribution *d) {
   if (d->n_bursts == 0) {
      d->n_bursts = 1;
      d->min_burst_pkts = d->max_burst_pkts = d->this_burst_pkts;
      }
   else {
      d->n_bursts += 1;
      if (d->this_burst_pkts < d->min_burst_pkts)
         d->min_burst_pkts = d->this_burst_pkts;
      if (d->this_burst_pkts > d->max_burst_pkts)
         d->max_burst_pkts = d->this_burst_pkts;
      }
   }
#endif

void bump_dist(struct distribution *d, double t_us)
{  /* t_us = pkt arrival time (us) */
   double x, T;
   int n = 0;

#if CHECK_DIST_SCALE
   if (d->Transform < 1 || d->Transform > DS_LOG_RESET ) {
      if (n_bad_t % 10 == 0)
         log_msg(LOG_ERR, 0, "bump_dist(1): flow %d, transform=%d, mask_params[0]=%d",
         flow_nbr(d->fp), d->Transform, d->mask_params[0]);
      ++n_bad_t;
      return;
      }
#endif

#if 0  // BURST_STATS
   this_ts = y;  /* d->last_us is in seconds, with precision ~ 1 us */
   if (d->total_count == 0) {
      d->this_burst_pkts = 1;
      d->burst_first_ts = d->last_ts = this_ts;
      return;
      }
   else {
      iat = this_ts - d->last_ts;
      if (iat <= BURST_GAP) {  /* Burst continues */
         d->this_burst_pkts += 1;
         d->last_ts = this_ts;  return;
         }
      else {  /* Start of next burst */
         end_burst(d);
         d->this_burst_pkts = 1;
         d->burst_first_ts = d->last_ts = this_ts;
         y = (double)iat;
         }  /* Count y (inter-burst gap) in d */
      }
#endif
#if 0  // approximate median algorithm
#define eta  0.01
#define sgn(x)  ( (x > 0) ? 1 : ( (x < 0) ? -1 : 0) )
   if (d->total_count == 1) d->median = x;  /* Use first iat */
   else d->median += eta * sgn(x - d->median);
   /* Algorithm from http://stackoverflow.com/questions/1058813/on-line-iterator-algorithms-for-estimating-statistical-median-mode-skewness */
#endif

   if (d->total_count == 0) x = 0.0;
   else x = t_us - d->last_us;  /* iat (s) */
   d->last_us = t_us;
   d->total_count += 1;  /* Counting started */

#if 0
   if ((d->Transform == DS_DYN_REQ || d->Transform == DS_DLOG_REQ)
         && d->n_values <= d->Buckets) {
      if (d->n_values == d->Buckets) {
            /* Too many actual values, start using bins */
         uint32_t values[MXBUCKETS+1];
         double R, N,D;
         int lo,hi, j;

         N = d->Buckets - 1;
         switch (d->Transform) {
         case DS_DYN_REQ:
            R = d->max_val == d->min_val ? 8  /* Bug fix, 25 Jun 02 */
               : d->max_val - d->min_val;  /* Leave room at both ends */
               /* max_val and min_val have been scaled */
            lo = d->min_val - 0.125*R;  /* May be negative */
            hi = d->max_val + 0.25*R;
	    scale_dist_limits(d, lo,hi);
            D = d->UpperLim - d->LowerLim;
            d->M = N/D;
            break;
         case DS_DLOG_REQ:
            lo = d->min_val*0.7;  /* Leave room at both ends */
            hi = d->max_val*3.0;;
	    scale_dist_limits(d, lo,hi);
            D = log10(d->UpperLim/d->LowerLim);
            d->M = N/D;
            break;
            }

#if CHECK_DIST_SCALE
 { char nbr[20], buf[2000], *p_p;
   log_msg(LOG_WARNING, 0,
      "DD(%d): lower=%d, upper=%d, min_val=%d, max_val=%d, scale=%d, "
         "Lower=%d, UpperLim=%d, UpperLimit=%d, HiDel=%d",
      d->selector, d->lowerspec,d->upperspec, d->min_val,d->max_val,
      d->Scale, d->LowerLimit, d->UpperLim, d->UpperLimit, d->HiDel);
   for (p_p = buf, j = 0; j != d->Buckets; ++j) {
      sprintf(nbr," %u", d->counts[j]);
      p_p = strmov(p_p, nbr);
      }
   *p_p = '\0';
   log_msg(LOG_WARNING, -1,  /* %% */
      "  counts=%s", buf);
   }
#endif
         memcpy(values, d->counts, sizeof(d->counts));
         memset(d->counts, 0, sizeof(d->counts));
         for (j = 0; j != d->Buckets; ++j) {
	   ix = (uint32_t)values[j];
               /* Assume uint32_t values to determine dynamic limits */
            if (ix <= d->LowerLim) n = 0;
            else if (ix > d->UpperLim) n = d->Buckets;
            else switch (d->Transform) {
            case DS_DYN_REQ:
               T = ((double)ix - d->LowerLim) * d->M;
               n = 1 + (int)floor(T);
               break;
            case DS_DLOG_REQ:
               T = log10((double)ix/d->LowerLim) * d->M;
               n = 1 + (int)floor(T);
               break;
               }
#if CHECK_DIST_SCALE
            if (n < 0 || n > d->Buckets || n > MXBUCKETS) {
               log_msg(LOG_ERR, 0, "bump_dist(2): flow %d, n=%d, transform=%d",
                  flow_nbr(d->fp), n, d->Transform);
               }
            else
#endif
               d->counts[n] += 1;
            }

         d->n_values = d->Buckets+1;
         }
      else {  /* d->n_values < d->Buckets */
	 // ix = yxpwr_of_10[d->Scale > 9 ? 9 : d->Scale];
	 // if (ix < d->min_val && ix >= d->lowerspec) d->min_val = ix;
	 // if (ix > d->max_val && ix <= d->upperspec) d->max_val = ix;

	 /* assign_uint32_t_to_c64(d->counts[d->n_values++], ix);
            Bug, showed up in uwisc 1s bitrate data.
            counts[] are 64-bit, don't force ix to bit32.
            Fixed by Nevil,    Sun 19 Mar 06 (IETF 65) */
         d->counts[d->n_values++] = t_us;
         d->non_empty = 1;
         return;
         }
      }
#endif

   // d->non_empty = 1;
   if (x <= d->LowerLim) {  /* Not a dynamic distrib */
      d->bin0_count += 1;
      return;
      }
   else if (x > d->UpperLim) {
      n = d->Buckets;
      }
   else switch (d->Transform) {
   case DS_LIN:
   case DS_DYN_REQ:
      T = (x - d->LowerLim) * d->M;
      n = 1 + (int)floor(T);
      break;
   case DS_LOG:
   case DS_DLOG_REQ:
      T = log10(x/d->LowerLim) * d->M;
      n = 1 + (int)floor(T);
      break;
      }
#if CHECK_DIST_SCALE
   if (n < 0 || n > d->Buckets || n > MXBUCKETS) {
      if (n_bad_n % 10 == 0)
         log_msg(LOG_ERR, 0, "bump_dist(3): flow %d, n=%d, transform=%d",
            flow_nbr(d->fp), n, d->Transform);
      ++n_bad_n;
      }
   else
#endif
      d->counts[n] += 1;
   }

double bin_edge(struct distribution *d, int bin) {
   switch (d->Transform) {
   case DS_LIN:
      return d->LowerLim + bin*d->step;
      break;
   case DS_LOG:
      return d->LowerLim * pow(10.0, bin*d->step);
      break;
      }
   return 0;  /* Keep compiler happy */
   }

int compute_distrib_stats(struct stream *sp) {
   int j, m_bin, nz,  cl,cr;
   uint32_t mode;
   double lifetime;
   struct distribution *d = sp->version == 4 ?
      sp->flow_info.fi4->d : sp->flow_info.fi6->d;

   mode = m_bin = nz_runs = 0;
   nz = d->bin0_count == 0 ? 1 : 0;
   for (j = 0; j != MXBUCKETS; ++j) {  /* Ignores oflo bin */
      if (d->counts[j] > mode) {
	 m_bin = j;  mode = d->counts[j];
         }
      if (d->counts[j] == 0) nz += 1;
      else if (nz != 0) {
	 if (nz > 1) nz_runs += 1;  /* Need two or more in a run */
         nz = 0;
         }
      }
   if (nz > 1) nz_runs += 1;  /* Run at top? */
   for (cl = 0, j = 0; j < m_bin-1; ++j) cl += d->counts[j];
   for (cr = 0, j = m_bin+2; j < MXBUCKETS; ++j) cr += d->counts[j];

   b0pc = d->bin0_count*100.0/d->total_count;
   modeiat = bin_edge(d, m_bin+1);
   if (m_bin > 0) mode += d->counts[m_bin-1];
   if (m_bin < MXBUCKETS-1) mode += d->counts[m_bin+1];
   skew = cr+cl == 0 ? 0.0 :(cr-cl)*100.0/(cl+cr);

   lifetime = sp->last_us - sp->first_us;  /* Classify -> IAT group */
   if (sp->totp < MXBUCKETS) {
          /* Too few pkts for reliable distribution stats */
      if (lifetime > 1800) {  /* long-lived */
	 if (modeiat > 3.0 && modeiat <= 4.0)
	    return IAT_stealth_3s_mode;
	 if (nz_runs <= 4) return IAT_stealth_spikes;
	 return IAT_stealth;
         }
      return IAT_short_lived;
      }
   else {
      if (sp->totp/lifetime > 5.0)  /* pkt/s */
	 return IAT_high_rate;
      if (b0pc > 10) return IAT_DOS;
      if (modeiat > 3.0 && modeiat <= 4.0) {
	 if (skew <= -10.0) return IAT_3_left;
         if (skew <= 10.0) return IAT_3_even;
	 return IAT_3_right;
         }
      }
   return IAT_Other;
   }

void write_distrib_info(FILE *f, struct stream *sp, struct distribution *d)
{
   fprintf(f, "#trace: %s\n", trace_name);
   fprintf(f, "#endtime: %s (UTC)\n", fmt_time(1, &last_ts));
   fprintf(f, "#meter: %s\n", meter_loc);
   fprintf(f, "#distrib: Transform=%u, Lower=%.3f, Upper=%.3f, Buckets=%u\n",
       d->Transform, d->LowerLim, d->UpperLim, d->Buckets);
   fprintf(f, "#title: inter-arrival time (s) " \
      "distribs for each one-way source\n");

#if 0  //BURST_STATS
   fprintf(f, "#   src, src_type,  pkt_av_len, sp->totp," \
      " d->n_bursts, d->min_burst_pkts, d->max_burst_pkts\n#\n");
#endif
   fprintf(f, "#attributes: src src_type  pktavlen totpkts" \
      "  ttlmode ttl2nd  portmode port2nd  flags keys  uT_count cfc_count" \
      "  first_s last_s  b0pc modeiat skew nzruns \n");
   fprintf(f, "#\n");
   }

void dump_source_info(FILE *f, struct stream *sp, int src_type,
   struct distribution *d)
{
   int j;  char *src = key_to_s(sp);
   int ttl_mode, ttl_2nd,  port_mode, port_2nd, uT_count, cfc_count;

#if 0 // BURST_STATS
   end_burst(d);
   fprintf(f, "%s %u  %u %u  %u %u %u ", src, src_type,
      sp->tb/sp->tp, sp->totp, 
      d->n_bursts, d->min_burst_pkts, d->max_burst_pkts);
#endif

   if (sp->version == 4) {
      struct v4_flow_info *fi4 = sp->flow_info.fi4;
      uT_count = fi4->uT_count;  cfc_count = fi4->cfc_count;
      ttl_mode = top2_uint8(&fi4->tn_ttl, &ttl_2nd);
      switch (sp->proto) {
      case PT_TCP:  
         port_mode = top2_uint16(&fi4->tn_tcp, &port_2nd);
         break;
      case PT_UDP:
         port_mode = top2_uint16(&fi4->tn_udp, &port_2nd);
         break;
      default:
 	 port_mode = port_2nd = -1;
         }
      }
   else {
      struct v6_flow_info *fi6 = sp->flow_info.fi6;
      uT_count = fi6->uT_count;  cfc_count = fi6->cfc_count;
      ttl_mode = top2_uint8(&fi6->tn_ttl, &ttl_2nd);
      switch (sp->proto) {
      case PT_TCP:  
         port_mode = top2_uint16(&fi6->tn_tcp, &port_2nd);
         break;
      case PT_UDP:
         port_mode = top2_uint16(&fi6->tn_udp, &port_2nd);
         break;
      default:
 	 port_mode = port_2nd = -1;
         }
      }

   fprintf(f,
      "%s %u  %llu %u  %d %d  %d %d  %02x %08x  %d %d  %.3f %.3f  %.2f %.2f %.2f %d ", 
      src, src_type, (sp->tb/sp->tp), sp->totp,
      ttl_mode, ttl_2nd,  port_mode, port_2nd, flags_mode, keys_mode,
      uT_count, cfc_count, sp->first_us, sp->last_us,
	   b0pc, modeiat, skew, nz_runs);

   fprintf(f, " %u", d->bin0_count);
   for (j = 0; j != d->Buckets; ++j)  fprintf(f, " %u", d->counts[j]);
   fprintf(f, "\n");
   }
uint32_t n_unanalysed_sources, n_unanalysed_pkts;
uint32_t    /* Stats for unanalysed sources */
   ua_pkts[MN_PKTS_FOR_DISCARD],  /* Nbr of packets */
   ua_src_addr[256],  /* Source addresses */
   ua_dst_addr[256],  /* Dest addresses */
   ua_ttl[256],       /* TTLs */
   ua_proto[256],     /* Protocols */
   ua_gt1prot;        /* More than one proto seen */
struct hashtable *ua_dst_ports, *ua_tcp_keys;

void init_ua_variables(void) {
   ua_gt1prot = n_unanalysed_sources = n_unanalysed_pkts = 0;

   memset(ua_pkts, 0, MN_PKTS_FOR_DISCARD*sizeof(uint32_t));
   memset(ua_src_addr, 0, 256*sizeof(uint32_t));
   memset(ua_dst_addr, 0, 256*sizeof(uint32_t));
   memset(ua_ttl, 0, 256*sizeof(uint32_t));
   memset(ua_proto, 0, 256*sizeof(uint32_t));

   ua_dst_ports = create_hashtable(5000, hashfromkey, equalkeys);
   ua_tcp_keys = create_hashtable(10000, hashfromkey, equalkeys);
   }

#define firstbyte(a)  ((ntohl(a) >> 24) & 0x000000FF)
#define secondbyte(a) ((ntohl(a) >> 16) & 0x000000FF)

void update_ua_stats(struct stream *sp) {
   int port = -1;

   if (n_unanalysed_sources == 0)  init_ua_variables();
   n_unanalysed_sources += 1;  n_unanalysed_pkts += sp->totp;

   if (sp->tp == 0) return;  /* Only seen 'from' pkts so far.  Thu, 1 Dec 11 */
   if (sp->tp > MN_PKTS_FOR_DISCARD)  /* Don't stop!  Tue, 22 Nov 11 */
      log_msg(LOG_WARNING, 0, "update_ua_stats(): tp = %d (too big)", sp->tb);
   else ua_pkts[sp->tp-1] += 1;
   ua_src_addr[firstbyte(sp->src_addr.a.v4)] += 1;
   ua_ttl[sp->ttl] += 1;
   if (sp->flow_info.fi4 != NULL) {
      if (sp->version == 4) {
         struct v4_flow_info *fi4 = sp->flow_info.fi4;
         if (fi4->tn_dst_addrs.nv == 0)
	    log_msg(LOG_ERR, 0, "fi4->tn_dst_addrs.nv == 0\n");
         ua_dst_addr[secondbyte(top_addrs4(&fi4->tn_dst_addrs))] += 1;
         ua_proto[top_uint8(&fi4->tn_proto)] += 1;
         if (fi4->tn_proto.nv == 1) {
	    if (fi4->tn_proto.values[0] == PT_TCP) {
	       port = top_uint16(&fi4->tn_tcp);
	       }
	    if (fi4->tn_proto.values[0] == PT_UDP) {
	       port = top_uint16(&fi4->tn_udp); 
	       }
	    }
	 else ua_gt1prot += 1;
         }
      }

   if (port >= 0) owt_ht_insert(ua_dst_ports, port);
   if (sp->tcp_key != 0)
      owt_ht_insert(ua_tcp_keys, sp->tcp_key); 
   }

/* We're using Christopher Clark's hashtable routines
      (by Christopher Clark, https://github.com/davidar/c-hashtable) */

struct value { int count; };  /* All our tables use uint32_t count as value */

#if HALF_WAY_STREAMS
/* Hash table routines for ports and tcp_keys .. */

DEFINE_HASHTABLE_INSERT(insert_some_addr, struct ip_address, struct value);
DEFINE_HASHTABLE_SEARCH(search_some_addr, struct ip_address, struct value);
DEFINE_HASHTABLE_REMOVE(remove_some_addr, struct ip_address, struct value);
DEFINE_HASHTABLE_ITERATOR_SEARCH(search_itr_some_addr, struct ip_address);

static unsigned int hashfromaddr(void *ky) {
   struct ip_address *a = ky;
   if (a->ver == 4) {
     //      return sp->src_addr.a.v4 ^ (4 * 97171);
      return (a->a.v4 >> 9) + a->a.v4 * 5;
      }
   else {  /* v6 stream */
      uint64_t hh = a->a.o.oct0 ^ a->a.o.oct1;
      return ((uint32_t *)&hh)[1] ^ ((uint32_t *)&hh)[0] ^ (6 * 96953);
      }
   }

static int equaladdrs(void *ky1, void *ky2) {
   struct ip_address *a = ky1, *b = ky2;
   if (a->ver == 4) {
      return
	 a->a.v4 == b->a.v4 &&
         a->ver == b->ver;
      }
   else {  /* v6 stream */
      return 
         a->a.o.oct1 == b->a.o.oct1 && a->a.o.oct0 == b->a.o.oct0 &&
         a->ver == b->ver;
      }

   }

static void addr_ht_insert(struct hashtable *h, struct ip_address *addr) {
   struct value *v, *found;
   struct ip_address *k = (struct ip_address *)malloc(sizeof(struct ip_address));
   if (k == NULL)
      log_msg(LOG_ERR, 111, "Out of memory allocating an ip_address");
   memcpy(k, addr, sizeof(struct ip_address));

   if ((found = search_some_addr(h, addr)) == NULL) {  /* Not found */
      v = (struct value *)malloc(sizeof(struct value));
      if (v == NULL)
         log_msg(LOG_ERR, 222, "Out of memory allocating a value");
      v->count = 1;
      if (!insert_some_addr(h,k,v))
         log_msg(LOG_ERR, 333, "Out of memory inserting a key");
      }
   else found->count += 1;
   }

static int addr_kv_cmp(const void *a, const void *b) {
   const struct addr_key_val *kva = a;
   const struct addr_key_val *kvb = b;
   if  (kvb->count > kva->count) return 1;
   else if (kvb->count == kva->count) return 0;
   return -1;
   }

struct addr_key_val *addr_get_kv( struct hashtable *h, int *length) {
   /* Return vector of (k,v) structs from ht, then destroy ht */
   struct hashtable_itr *itr = hashtable_iterator(h);
   int j, len = hashtable_count(h);
   struct addr_key *k;  struct value *v;
   struct addr_key_val *kv = NULL;

   *length = len;
   if (len > 0) {
      kv = (struct addr_key_val *)malloc(len*sizeof(struct addr_key_val));
      if (kv != NULL) {
         j = 0;  do {
	    k = hashtable_iterator_key(itr);
	    v = hashtable_iterator_value(itr);
 	    kv[j].key = k->key;  kv[j].count = v->count;
	    j += 1;
            } while (hashtable_iterator_advance(itr));
         }
      qsort(kv, len, sizeof(struct addr_key_val), addr_kv_cmp);
      }
   free(itr);
   hashtable_destroy(h,1); /* Second arg says "free(value)" */
 
   return kv;
   }
#endif

/* Hash table routines for ports and tcp_keys .. */

DEFINE_HASHTABLE_INSERT(insert_some, struct key, struct value);
DEFINE_HASHTABLE_SEARCH(search_some, struct key, struct value);
DEFINE_HASHTABLE_REMOVE(remove_some, struct key, struct value);
DEFINE_HASHTABLE_ITERATOR_SEARCH(search_itr_some, struct key);

static unsigned int hashfromkey(void *ky) {
   struct key *k = (struct key *)ky;
   return (k->key >> 9) + k->key * 5; 
   } 

static int equalkeys(void *ky1, void *ky2) {
   struct key *k1 = (struct key *)ky1,
      *k2 = (struct key *)ky2;
   return (k1->key == k2->key);
   }

static void owt_ht_insert(struct hashtable *h, uint32_t tk) {
   struct value *v, *found;
   struct key *k = (struct key *)malloc(sizeof(struct key));
   if (k == NULL)
      log_msg(LOG_ERR, 111, "Out of memory allocating a key");
   k->key = tk;

   if ((found = search_some(h,k)) == NULL) {  /* Not found */
      v = (struct value *)malloc(sizeof(struct value));
      if (v == NULL)
         log_msg(LOG_ERR, 222, "Out of memory allocating a value");
      v->count = 1;
      if (!insert_some(h,k,v))
         log_msg(LOG_ERR, 333, "Out of memory allocating a key");
      }
   else found->count += 1;
   }

static int kv_cmp(const void *a, const void *b) {
   const struct key_val *kva = a;
   const struct key_val *kvb = b;
   if  (kvb->count > kva->count) return 1;
   else if (kvb->count == kva->count) return 0;
   return -1;
   }

struct key_val *get_kv( struct hashtable *h, int *length) {
   /* Return vector of (k,v) structs from ht, then destroy ht */
   struct hashtable_itr *itr = hashtable_iterator(h);
   int j, len = hashtable_count(h);
   struct key *k;  struct value *v;
   struct key_val *kv = NULL;

   *length = len;
   if (len > 0) {
      kv = (struct key_val *)malloc(len*sizeof(struct key_val));
      if (kv != NULL) {
         j = 0;  do {
	    k = hashtable_iterator_key(itr);
	    v = hashtable_iterator_value(itr);
 	    kv[j].key = k->key;  kv[j].count = v->count;
	    j += 1;
            } while (hashtable_iterator_advance(itr));
         }
      qsort(kv, len, sizeof(struct key_val), kv_cmp);
      }
   free(itr);
   hashtable_destroy(h,1); /* Second arg says "free(value)" */
 
   return kv;
   }

/* ----- End of hash table routines ----- */
