/* 1541, Thu 17 Mar 11 (PDT)
   1552, Thu 17 Jun 10 (NZST)
   1955, Thu 22 Apr 10 (PDT)
   1330, Fri  1 Feb 08 (NZDT)

   owt_watcher.c:  Packet and stream routines for owtmon

   iatmon/owtmon: One-Way Traffic Monitor testbed
   Copyright (C) 2009-2011 by Nevil Brownlee, U Auckland | CAIDA | WAND

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "owt.h"

/* Declarations for SRI code to generate Conficker port numbers (in conf-sri.c) */
typedef union {
   int16_t s16[8];  uint16_t u16[8];  int32_t s32[4];
   } result_t;
int portgen(int ip, result_t *res, int week);

uint32_t date_seed;

void set_date_seed(uint32_t ts) {
   /* Emile's python code ...
      date_seed = ((((uint64_t)ts - 345600) * 1861606989) >> 32) >> 18; */
   /* Nevil's C code: following line seems to be equivalent to
      date_seed = ((uint64_t)ts - 4*24*3600)/(7*24*3600); */
   date_seed = (uint32_t)(((((uint64_t)ts - 345600) * 1861606989) >> 32) >> 18);
   //d printf("   l_t_s=%u, date_seed=%x\n", ts, date_seed);
   }

void swap_ip_addresses(struct ip_address *s, struct ip_address *d) {
   struct ip_address temp;
   memcpy(&temp, s, sizeof(struct ip_address));
   memcpy(s, d, sizeof(struct ip_address));
   memcpy(d, &temp, sizeof(struct ip_address));
   }

static char const *bT_DHT_val = "d1:ad2:id20:";  /* 12 bytes 0..11 */
static char const *uT_proto_val =  /* 16 bytes 12..27 */
   "\x7f\xff\xff\xff\xab\x02\x04\x00\x01\x00\x00\x00\x08\x00\x00\x00";

#define test_ut \
   sp->uTorrent = \
      (hl >= dx+8+0+12 && \
         (memcmp(&hdr[dx+8+0], bT_DHT_val, 12) == 0)) || \
      (hl >= dx+8+12+16 && \
       (memcmp(&hdr[dx+8+12], uT_proto_val, 16) == 0));

uint32_t tcp_key(int dx, unsigned char *hdr, int hl) { 
   uint32_t val;  int x;

   if (hl < dx+16) return 0;
   if ((hdr[dx+13] & 0x02) == 0)  /* SYN flag not present */
      return 0;
   val = ntohs(*(uint16_t *)&hdr[dx+14]);  /* Window size */

   if ((hdr[dx+12] >> 4) == 5) return val;  /* No options */

   x = dx+20;
   for (;;) { 
      switch (hdr[x]) { 
      case 0:  /* End of option list */ 
         return val; 
      case 1:  /* No-op */ 
         x += 1; 
         break;
      case 2:  /* MSS */ 
         val |= (ntohs(*(uint16_t *)&hdr[x+2])-600) << 16;  /* 10 bits */
         break;
      case 3:  /* Window scale */ 
         val |= ntohs(*(uint16_t *)&hdr[x+2]) << 28;  /* Top 4 bits */
         break;
      case 4:  /* SACK permitted */
         val |= 0x08000;
         break;
      case 8:  /* Timestamp */
         val |= 0x04000000;
         break;
         }
      if (x+1 >= hl || hdr[x+1] == 0) break; 
      x += hdr[x+1]; 
      if (x >= hl) break; 
      } 
   return val; 
   } 

int get_v4key(struct stream **r_sp, int *r_fwd, int hl, unsigned char *hdr)
{  /* Code from pkt_extract() in NeTraMet's flowhash.c */
   struct stream *sp;
   int ver, dx, frag;
   struct ip_address *saddr, *daddr;  uint16_t sport, dport;
   int sa_local, da_local, fwd;

   if ((ver = hdr[0] >> 4) == 4) {
      if (hl < 12+IP4_ADDR_LEN*2) {
         *r_sp = NULL;
         return SKEY_NOPEERADDR;  /* Didn't get the peer addresses! */
         }
      if (frag = (ntohs(*(uint16_t *)&hdr[6]) & 0x1FFF) != 0) {
 	 *r_sp = NULL;
         return SKEY_FRAGPKT;  /* Non-first fragment */
         }

      if ((sp = get_stream(4)) == NULL) return SKEY_NOSPACE;
      sp->version = 4;
      sp->proto = hdr[9]; /* Protocol Type */
      dx = (hdr[0] & 0x0F) << 2;  /* IPv4 header length in 32-bit units */
      sp->tcp_flags = 0;

      saddr = &sp->src_addr;  daddr = &sp->dst_addr;
      saddr->ver = daddr->ver = 4;  saddr->len = daddr->len = 4;
      saddr->a.v4 = *(uint32_t *)&hdr[12];  /* IPv4 source */
      daddr->a.v4 = *(uint32_t *)&hdr[16];  /* IPv4 dest */
      sa_local = addr_is_local(4, saddr);
      da_local = addr_is_local(4, daddr);
      if (sa_local && !da_local) fwd = 0;
      else if (!sa_local && da_local) fwd = 1;
      else {
         free_stream(sp);  *r_sp = NULL;
         return SKEY_NOLOCALADDR;
         }
      if (!fwd) swap_ip_addresses(saddr, daddr);

      switch (sp->proto) {
      case PT_TCP:
#if HALF_WAY_STREAMS
 	 sp->tcp_seq = ntohl(*(uint32_t *)&hdr[dx+4]);
 	 sp->tcp_ack = ntohl(*(uint32_t *)&hdr[dx+8]);
	    /* We only want to see whether seq or ack is increasing,
               so we don't need to check the ACK flag */
#endif
         sp->tcp_flags = hl >= dx+14 ? hdr[dx+13] : 128;

      case PT_UDP:
         if (hl < dx+4) {
  	    free_stream(sp);  *r_sp = NULL;
            return SKEY_NOTRANSADDR;  /* Didn't get the IP ports! */
            }
         else {
            sport = ntohs(*(uint16_t *)&hdr[dx]);  /* Source port */
            dport = ntohs(*(uint16_t *)&hdr[dx+2]);  /* Dest port */
	    }

         if (fwd) {
            sp->src_port = sport;  sp->dst_port = dport;
            }
         else {
            sp->src_port = dport;  sp->dst_port = sport;
            }
         //      printf("*** src=%08x, dst=%08x, fwd=%d\n",  //p
         //	     sd4p->src_addr, sd4p->dst_addr, fwd);
         break;
      case PT_ICMP:
         sp->src_port = 0;    /* Type field */
         sp->dst_port = ntohs(*(uint16_t *)&hdr[dx]);  /* Type|Code */
         break;
      default:
         sp->src_port = sp->dst_port = 0;
         break;
         }
      //d  printf("get_v4key(1): pktp=%0x, saddr=%x, daddr=%x, sp=%x\n",
      //d     pktp, saddr, daddr, sp);


      sp->ip_len = ntohs(*(uint16_t *)&hdr[2]);
      sp->ttl = hdr[8];
      sp->uTorrent = 0;  sp->tcp_key = 0;
      if (sp->proto == PT_TCP) {
	 sp->tcp_key = tcp_key(dx, hdr, hl);
         }
      else if (sp->proto == PT_UDP) {
         test_ut;
         }
      *r_fwd = fwd;
      *r_sp = sp;  /* Return stream (with ptrs to pkt_data and sd_tuple_v4) */
      return SKEY_OK;
      }
   else return SKEY_BADIPHDR;
   }

int find_v6_type(int *prot, int *hx, const uint8_t *p, const int hl)
{
   int nh, nx;

   nh = p[6];  nx = 40;
   for (;;) {
      switch (nh) {
      case PT_6FRAG:    /* Fragment */
      case PT_6ESP:     /* Encapsulating Security Payload (ESP) RFC 2406 */
      case PT_6NONH:    /* No next header */
         *prot = nh;
         return 0;  /* Can't analyse this packet */

      case PT_6HOP:     /* Hop-by-hop options */
      case PT_6ROUT:    /* Routing (Type 0) */
      case PT_6DEST:    /* Destination options */
      case PT_6AH:      /* Authentication (AH) RFC 2402 */
      case PT_6MOBILE:  /* IPv6 Mobility Header */
	 break;  /* Keep looking */

      default:
         *prot = nh;  *hx = nx;
         return 1;  /* Can see into this one */
         }
      if (nx >= hl) {
	 *prot = -1;  /* Not enough header bytes to get Next Header */
         return 0;
         }
      nh = p[nx];
      nx += (p[nx+1]+1) * 8;
      }
   }

int get_v6key(struct stream **r_sp, int *r_fwd, int hl, unsigned char *hdr)
{  /* Code from pkt_extract() in NeTraMet's flowhash.c */
   struct stream *sp;
   struct ip_address *saddr, *daddr;  uint16_t sport, dport;
   int ver, dx, prot;
   int sa_local, da_local, fwd;

   if ((ver = hdr[0] >> 4) == 6) {
      if (hl < 8+IP6_ADDR_LEN*2) {
         *r_sp = NULL;
         return SKEY_NOPEERADDR;  /* Didn't get the peer addresses! */
         }
      if (!find_v6_type(&prot,&dx, hdr,hl)) {
         *r_sp = NULL;
         if (prot < 0)  /* Not enough bytes to determine protocol */
  	    return SKEY_V6SMALLSNAP;
         else return SKEY_V6OPAQUE;  /* Can't see into packet */
         }

      if ((sp = get_stream(6)) == NULL) return SKEY_NOSPACE;
      sp->version = 6;
      saddr = &sp->src_addr;  daddr = &sp->dst_addr;
      saddr->ver = daddr->ver = 6;  saddr->len = daddr->len = 16;
      memcpy(saddr->a.a, &hdr[8], IP6_ADDR_LEN);  /* IPv6 source */
      memcpy(daddr->a.a, &hdr[24],IP6_ADDR_LEN);  /* IPv6 dest */
      sa_local = addr_is_local(6, saddr);
      da_local = addr_is_local(6, daddr);
      if (sa_local && !da_local) fwd = 0;
      else if (!sa_local && da_local) fwd = 1;
      else {
    	 free_stream(sp);  *r_sp = NULL;
         return SKEY_NOLOCALADDR;
         }
      if (!fwd) swap_ip_addresses(saddr, daddr);

      switch (prot) {
      case PT_TCP:
#if HALF_WAY_STREAMS
 	 sp->tcp_seq = ntohl(*(uint32_t *)&hdr[dx+4]);
 	 sp->tcp_ack = ntohl(*(uint32_t *)&hdr[dx+8]);
	    /* We only want to see whether seq or ack is increasing,
               so we don't need to check the ACK flag */
#endif
         sp->tcp_flags = hl >= dx+14 ? hdr[dx+13] : 128;
      case PT_UDP:
         if (hl < dx+4) {
            free_stream(sp);  *r_sp = NULL;
            return SKEY_NOTRANSADDR;  /* Didn't get the IP ports! */
            }
         else {
            sport = ntohs(*(uint16_t *)&hdr[dx]);  /* Source port */
            dport = ntohs(*(uint16_t *)&hdr[dx+2]);  /* Dest port */
            }
         if (fwd) {
            sp->src_port = sport;  sp->dst_port = dport;
            }
         else {
            sp->src_port = dport;  sp->dst_port = sport;
            }
         break;
      case PT_6ICMP:
         sp->src_port = 0;    /* Type field */
         sp->dst_port = ntohs(*(uint16_t *)&hdr[dx]);  /* Type|Code */
         break;
      default:
         sp->src_port = sp->dst_port = 0;
         break;
         }

      sp->ip_len = 40 + ntohs(*(uint16_t *)&hdr[4]);
         /* Header + payload (includes extension headers) */
      sp->ttl = hdr[7];

      if (sp->proto == PT_TCP) {
	 sp->tcp_key = tcp_key(dx, hdr, hl);
         }
      else if (sp->proto == PT_UDP) { 
         test_ut;
         }

      *r_fwd = fwd;
      *r_sp = sp;  /* Return stream (with ptrs to pkt_data and sd_tuple_v6) */
      return SKEY_OK;
      }
   else return SKEY_BADIPHDR;  /* Wasn't a v6 packet! */
   }

void print_stream(struct stream *sp) {
   if (sp->version != 4) printf("stream: type=%d <<<\n", sp->version);
   else {
      uint8_t *a = sp->src_addr.a.a;
      printf("%u.%u.%u.%u", a[0],a[1],a[2],a[3]);
      a = sp->dst_addr.a.a;
      printf("->%u.%u.%u.%u", a[0],a[1],a[2],a[3]);
      }
   }

void check_str_nulls(struct stream_queue *sq, struct stream *sp);
int check_stream_chains(char const *msg, struct stream_queue *sq);

void enqueue_stream_status(int type, struct stream *sp, uint32_t now_ts) {
   struct opq_entry *oqe;
   t_semaphore_down(&opq_empty);  /* Wait for an empty entry */
   oqe = oqe_for_stream(sp);
   oqe->type = type;
   opq_entry_q_lock(&opq);
   opq_entry_push(&opq, oqe);
   opq_entry_q_unlock(&opq);
   sp->last_report_sec = now_ts;
   t_semaphore_up(&opq_ready);  /* Signal a full entry */
   }

void enqueue_stream_txt(char *msg) {
   struct opq_entry *oqe;
   t_semaphore_down(&opq_empty);  /* Wait for an empty entry */
   oqe = oqe_for_stream(NULL);
   oqe->type = OQ_TXTMSG;  oqe->msg = msg;
   oqe->ts = g_last_ts;
   opq_entry_q_lock(&opq);
   opq_entry_push(&opq, oqe);
   opq_entry_q_unlock(&opq);
   t_semaphore_up(&opq_ready);  /* Signal a full entry */
   }

static int rake_idle_streams(struct flow_watcher_control *fw, uint32_t now_s) {
   struct stream_queue *sq = &fw->sq;
   double min_idle_us;
   struct stream *sp;
   int n, keep, examined = 0, freed = 0;

#if SC_DEBUG >= 2
   fprintf(stderr, "  === check_timeouts(%d), ts=%u\n", fw->id, now_ts);
#endif
   min_idle_us = (now_s - TOO_MANY_TIMEOUT)*1000000.0;
#if SC_DEBUG >= 1
   check_stream_chains("--- starting check_stream_timeouts", sq);
#endif
   stream_q_lock(sq);
   for (n = sq->length; n != 0;  --n) {  /* Don't go round more than once */
      sp = stream_shift(sq);  /* Remove stream from head */
      if (sp->last_us > min_idle_us) {
 	 stream_unshift(sq, sp);  /* Put it back! */
         break;  /* Hasn't been queued long enough; stop checking */
         }
      else {  /* Could be inactive, can we recover it? */
	 examined += 1;
	 keep = sp->totp > MN_PKTS_FOR_DISCARD;
         }
      if (keep) stream_push(sq, sp);  /* Add stream to tail */
         /* We won't waste time tesing it until it drifts back
            to the head of the queue.  That means we no longer
            have the queue in (approximnate) LastTime order */
      else {  /* Discard this stream */
	 update_ua_stats(sp);  /* Un-analysed source info */
         update_summary_info(sp); 
         owt_ht_delete(fw->sh_table, sp);  /* Remove from hash table */
	 sp->next = sp->prev = (struct stream *)0x55555555;
         free_stream(sp);
	 freed += 1;
	 if (freed == MN_STREAMS_TO_FREE) break;
	 //	 check_stream_chains("--- after timed out", sq);
         }
      }
#if SC_DEBUG >= 2
   n = stream_space.mx_objects - stream_space.free_queue.length;
   fprintf(stderr, "   ==== streams: %d examined, %d freed.  %d active sv4s\n",
      examined, freed, n);
#endif

   stream_q_unlock(sq);
   return freed;
   }

void check_stream_timeouts(struct flow_watcher_control *fw, uint32_t now_ts) {
   struct stream_queue *sq = &fw->sq;
   double min_idle_us2, min_idle_us1, last_t_us, av_int;
   struct stream *sp;
   int keep, terminated, n, ow_terminated;
   int counted = 0, freed = 0;

#if SC_DEBUG >= 2
   fprintf(stderr, "  === check_timeouts(%d), ts=%u\n", fw->id, now_ts);
#endif
   last_t_us = now_ts*1000000.0;
   min_idle_us1 = (now_ts - ONE_WAY_TIMEOUT)*1000000.0;
   min_idle_us2 = (now_ts - TWO_WAY_TIMEOUT)*1000000.0;
   ow_terminated = terminated = 0;
#if SC_DEBUG >= 1
   check_stream_chains("--- starting check_stream_timeouts", sq);
#endif
   stream_q_lock(sq);
   for (n = sq->length; n != 0;  --n) {  /* Don't go round more than once */
      counted += 1;
      sp = stream_shift(sq);  /* Remove stream from head */
      if (sp->last_us > (sp->timeout == STO_ONE_WAY ?
            min_idle_us1 : min_idle_us2) ) {
 	 stream_unshift(sq, sp);  /* Put it back! */
         break;  /* Hasn't been queued long enough; stop checking */
         }
      else {  /* Could be inactive, can we recover it? */
         if (sp->totp == 1)
            keep = 0;  /* Only one packet seen */
         else {
	    av_int = (sp->last_us - sp->first_us)/
               (double)(sp->totp - 1);
            keep = (last_t_us - sp->last_us)  /* Observed idle us */
	       < av_int*STO_MULTIPLIER;  /* Current timeout estimate */
            }
         }
      if (keep) stream_push(sq, sp);  /* Add stream to tail */
         /* We won't waste time tesing it until it drifts back
            to the head of the queue.  That means we no longer
            have the queue in (approximnate) LastTime order */
      else {  /* Timed-out stream */
         if (sp->state != SS_TWO_WAY)
            stream_push(sq, sp);  /* Add stream to tail */
   	 else {  /* Two-way or DELETE */
            terminated += 1;
            update_summary_info(sp); 
            sp->state = SS_TIMEOUT;
            owt_ht_delete(fw->sh_table, sp);  /* Remove from hash table */
	    sp->next = sp->prev = (struct stream *)0x55555555;
            free_stream(sp);
	    freed += 1;
            }
	 //	 check_stream_chains("--- after timed out", sq);
         }
      }
#if SC_DEBUG >= 2
   n = stream_space.mx_objects - stream_space.free_queue.length;
   fprintf(stderr, "   ==== %d streams terminated, %d freed, %d ignored, %d counted.  %d active sv4s\n",
      terminated, freed, ignored, counted, n);
#endif

   stream_q_unlock(sq);
   }

uint32_t swp_now_ts;
int get_stream_with_pkts(struct stream *sp)
{  /* Return 0 to continue, 1 to stop walking */
   if (sp->state == SS_ONE_WAY)  /* One-way stream */
      enqueue_stream_status(OQ_END_OF_FILE, sp, swp_now_ts);
   return 0;
   }

void output_active_streams(struct stream_queue *sq, uint32_t now_ts) {
   stream_q_lock(sq);
   check_stream_chains("--- starting output_active_src_hosts",sq);
   swp_now_ts = now_ts;
   stream_walk_up(sq, get_stream_with_pkts);
   check_stream_chains("+++ leaving output_active_src_hosts", sq);
   stream_q_unlock(sq);
   }

void check_str_nulls(struct stream_queue *sq, struct stream *sp) {
   if (sp->next == NULL && sp != sq->tail) {
      printf("null sp.next and sp != sp tail ***\n");
      quack(991);
      }
   if (sp->prev == NULL && sp != sq->head) {
      printf("null sp.prev and sp != s head ***\n");
      quack(992);
      }
   }

static char ftb[32];
char *fmt_time(int utc_time, uint32_t *ti) {  /* From NeTraMet */
#if SIZEOF_TIME_T == 8
   uint64_t lti = *ti;
#else
   uint32_t lti = *ti;
#endif
   if (!utc_time) {
      char *ts = ctime((time_t *)&lti);
      strncpy(&ftb[0],  &ts[11], 9);  /* 17:31:42_ */
      strncpy(&ftb[9],  &ts[0],  4);  /* Thu_ */
      strncpy(&ftb[13], &ts[8],  3);  /* 23_  */
      strncpy(&ftb[16], &ts[4],  4);  /* Sep_ */
      strncpy(&ftb[20], &ts[20], 4);  /* 1997 */
      ftb[24] = '\0';
      }
   else {
      struct tm *t = gmtime((time_t *)&lti);
      char const *dow[] = {
         "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
      char const *moy[] = {
         "Jan", "Feb", "Mar", "Apr", "May", "Jun",
         "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
      sprintf(ftb, "%02d:%02d:%02d %s %2d %s %4d",
         t->tm_hour,t->tm_min,t->tm_sec, dow[t->tm_wday],
	 t->tm_mday, moy[t->tm_mon], t->tm_year+1900);
      }
   return ftb;
   }   

static uint32_t last_timerec_sec;
static void output_time_record(uint32_t now_ts) {
   struct stream_txt *st = get_stream_txt_object(&stream_txt_space);
   if (st != NULL) {
      sprintf(st->msg, "#Time: %s  %s  Flows from %lu to %lu",
         fmt_time(1, &now_ts), trace_name,
         (unsigned long)last_timerec_sec, (unsigned long)now_ts);
      enqueue_stream_txt(st->msg);
      }
   last_timerec_sec = now_ts;
   }

static void fdf_hdr_string(char *buf) {
   sprintf(buf, "##%s: -c%u %s -s%u  starting at %s",
      version, time_rec_interval, trace_name,
      stream_space.mx_objects, fmt_time(1, &start_ts));
   }

static void output_fdf_hdr(void) {
   struct stream_txt *st = get_stream_txt_object(&stream_txt_space);
   if (st != NULL) {
      fdf_hdr_string(st->msg);
      enqueue_stream_txt(st->msg);
      }
   }

char *owt_file_header(void) {
   char *msg = malloc(MXSTRMTXTLEN);
   fdf_hdr_string(msg);
   return msg;
   }

void update_interface_stats(uint32_t now_ts) {
   /* Called every STATS_INTERVAL */
   int pd = pkt_drops();
   struct stream_txt *st = get_stream_txt_object(&stream_txt_space);

   if (st != NULL) {
      uint32_t j, av_pkt_count, n_seconds, mx_pkt_rate;
      struct if_info *iip;

      av_pkt_count = n_seconds = mx_pkt_rate = 0;
      mutex_lock("uis", &if_info_mutex);
      for (j = 0; j != n_interfaces; ++j) {
         iip = &if_info[j];
         av_pkt_count += iip->av_pkt_count;
         if (iip->n_seconds > n_seconds) n_seconds = iip->n_seconds;
         if (iip->mx_pkt_rate > mx_pkt_rate) mx_pkt_rate = iip->mx_pkt_rate;
         iip->mx_pkt_rate = iip->av_pkt_count = iip->n_seconds = 0;
         }
      mutex_unlock("uis", &if_info_mutex);

      double apc = av_pkt_count == 0 ? 1.0 : av_pkt_count;
      double r_hc_searches = n_hc_searches == 0 ? 1.0 : (double)n_hc_searches;
      double acl = (double)av_cl_count/r_hc_searches;
      uint32_t aps = (n_seconds == 0) ? av_pkt_count
         : (av_pkt_count*10+5)/(n_seconds*10);
      sprintf(st->msg, "#Stats: %u  aps=%u mps=%u lpk=%u lpp=%.3f" \
            "  acl=%.3f mst=%u fi4=%.1f fi6=%.2f" \
	    "  nip=%u nfr=%u nts=%u  mql=%u",
         now_ts,  aps, mx_pkt_rate, pd, (double)pd*100.0/apc,
            acl, mx_active_streams,
	    (double)mx_active_v4_flow_infos*100.0/mx_active_streams,
	    (double)mx_active_v6_flow_infos*100.0/mx_active_streams,
         n_not_ip, n_frags, n_too_small, mx_active_opq_entries);
      enqueue_stream_txt(st->msg);
      }
   n_not_ip = n_frags = n_too_small = n_no_local_address =
      av_cl_count = n_hc_searches = n_v6_opaque =
      mx_active_streams = mx_active_v4_flow_infos = mx_active_v6_flow_infos =
      mx_active_opq_entries = 0;

#if MKSTATSCOUNTS  /* Move on to next interval */
   active_sources_this_interval = 
      active_sources_per_stats_interval[stats_interval_nbr];
   unique_hosts_this_interval =
      unique_hosts_per_stats_interval[stats_interval_nbr];
   //x   pkts_seen_this_interval = 
   //x      pkts_seen_per_stats_interval[stats_interval_nbr];
   ++stats_interval_nbr;
#endif

#if MONITOR_PKTD_STATE
   gather_stream_pkt_stats();
#endif
   }

static void each_time_rec(uint32_t now_ts) {
   /* Called every TIME_INTERVAL */
   output_time_record(now_ts);
#if STREAMTIMEOUTS  /* For trace file, just write summary at EOF */
   struct stream_txt *st = get_stream_txt_object(&stream_txt_space);
   if (st == NULL)
      log_msg(LOG_ERR, 1, "each_time_rec() couldn't get stream_txt");
   sprintf(st->msg, "#Record: time to write a sum file!");
   enqueue_stream_txt(st->msg);
#endif
   }

int fw_id = 0;
void start_flow_watcher(struct flow_watcher_control *fw, 
      uint32_t pw_q_sz, int ht_size) {
   uint32_t j;
   fw->id = fw_id;  fw_id += 1;
#if STARTUP_DEBUG
      printf("start_flow_watcher(1): pw_q_sz=%d, ht_size=%d\n",  //dd
   	  pw_q_sz, ht_size);  //dd
#endif
   fw->sh_table = owt_ht_create(ht_size);  /* Stream hash table */
   if (fw->sh_table == NULL)
      log_msg(LOG_ERR, 1, "Couldn't create hash table <<<");

   init_stream_queue(&fw->ispq);
   init_stream_queue(&fw->sq);
   t_semaphore_init(&fw->ispq_ready);
   t_semaphore_decrement(&fw->ispq_ready);  /* Thread blocked (v=0) */
   t_semaphore_init(&fw->ispq_empty);
   for (j = 0; j != pw_q_sz-1; ++j)
      t_semaphore_up(&fw->ispq_empty);  /* v=(nbr entries avail) */

   mutex_lock("sfw", &n_fw_mutex);
   n_fw_running += 1;
   mutex_unlock("sfw", &n_fw_mutex);

   if ((fw->eof_sp = get_stream(6)) == NULL)
      log_msg(LOG_ERR, 1, "Couldn't get stream in start_flow_watcvher()\n");
#if STARTUP_DEBUG
   printf("start_fw() about to call pthread_create()\n"); fflush(stdout);
#endif
   if (pthread_create(&fw->flow_thread, NULL, flow_watcher_process, fw)) {
      fprintf(stderr, "Error creating flow-watch thread <<<\n");
      exit(41);
      }
#if STARTUP_DEBUG
   printf("leaving start_fw +-+\n"); fflush(stdout);
#endif
   }

int n_flow_watchers, fw_next_st_to_check;

FILE *tcp_keys_f;  //tcpk

void start_meter(char *interfaces[], uint32_t n_ifs,
   uint32_t mx_events, uint32_t mx_opq_entries, uint32_t mx_streams,
   uint32_t pw_q_sz, uint32_t mx_stream_txts, uint32_t mx_flow_watchers,
   uint32_t mx_source_types, uint32_t hours_reqd)
{
   struct event *ep;
   char name[250], *np;
   int err, r;  uint32_t j, ht_size;
   struct if_info *iip;

   //   tcp_keys_f = fopen("tcp_keys.txt", "w");  //tcpk

#if STARTUP_DEBUG
   printf("entering start_meter() +-+\n");  fflush(stdout);
#endif
   sprintf(name, "One-Way Traffic Monitor v%s", METER_VERSION);
   version = malloc(strlen(name)+1);
   strcpy(version, name);
   n_source_types = mx_source_types;

#if MKSTATSCOUNTS
#define N_INTERVALS  (traces_to_process*STATS_INTERVALS_PER_HOUR +1)
   active_sources_per_stats_interval =  /* Allocate per-src_type rows */
      calloc(N_INTERVALS, sizeof(uint32_t *));
      /* +1 because we see a few packets recorded after the hour change! */
   unique_hosts_per_stats_interval =
      calloc(N_INTERVALS,  sizeof(uint32_t *));
   //x   pkts_seen_per_stats_interval =
   //x      calloc(N_INTERVALS, sizeof(uint32_t *));
   for (j = 0; j != N_INTERVALS; ++j) {
      /* Allocate per-minute rows */
      active_sources_per_stats_interval[j] =
	calloc(n_source_types, sizeof(uint32_t));
      unique_hosts_per_stats_interval[j] =
	calloc(n_source_types, sizeof(uint32_t *));
      //x      pkts_seen_per_stats_interval[j] =
      //x	calloc(n_source_types, sizeof(uint32_t *));
      }
   active_sources_this_interval = active_sources_per_stats_interval[0];
   unique_hosts_this_interval = unique_hosts_per_stats_interval[0];
   //x   pkts_seen_this_interval = pkts_seen_per_stats_interval[0];
   stats_interval_nbr = 1;  /* 1-org interval nbrs */
#endif
   n_unanalysed_sources = n_unanalysed_pkts = 0;
#if MKDESTCOUNTS
   dest_addrs_per_source_type = malloc(n_source_types*sizeof(uint32_t *));
   for (j = 0; j != n_source_types; ++j) {
      dest_addrs_per_source_type[j] = calloc(NDESTBINS, sizeof(uint32_t));
      }
#endif

   request_shutdown = count = 0;
   owt_start(mx_events, mx_opq_entries, mx_streams,
      mx_stream_txts, mx_flow_watchers);
   last_timerec_sec = 0;  uname(&node);

   mx_active_streams = mx_active_v4_flow_infos = mx_active_v6_flow_infos =
      mx_active_opq_entries = 0;
   n_interfaces = n_ifs;
   for (err = j = 0; j != n_interfaces; ++j) {
      iip = &if_info[j];
      iip->if_name = interfaces[j];
      if ((r = init_trace(iip)) != 0)  /* Do this before scheduling events! */
	 err = 1;
      iip->if_nbr = j;  iip->n_pkts = 0;
      iip->mx_pkt_rate = iip->av_pkt_count = iip->n_seconds = 0;
      }
   if (err) exit(99);
   n_not_ip = n_frags = n_too_small = n_v6_opaque =
      intervals_this_sum_file = 0;

   sprintf(name, "%s", interfaces[0]);
   np = name + strlen(name);
   for (j = 1; j < n_ifs; ++j) {
      sprintf(np, ", %s", interfaces[j]);
      np = name + strlen(name);
      }
   sprintf(np, " on %s", node.nodename);
   trace_name = malloc(strlen(name)+1);
   strcpy(trace_name, name);

   t_semaphore_init(&opq_ready);
   t_semaphore_decrement(&opq_ready);  /* Thread blocked (v=0) */
   t_semaphore_init(&opq_empty);
   for (j = 0; j != opq_entry_space.mx_objects-1; ++j)
       t_semaphore_up(&opq_empty);  /* v=(nbr entries avail) */

   output_fdf_hdr();  /* Output .fdf header record(s) */

   ep = get_event(S_TIME_EVENT, 1);    /* Every 1s */ 
   schedule_event(ep, start_ts+3);  /* First time 3s from now */

   ep = get_event(STATS_EVENT, STATS_RECORDING_INTERVAL);
   schedule_event(ep, start_ts+4);

   ep = get_event(TIME_EVENT, time_rec_interval);
   schedule_event(ep, (start_ts+time_rec_interval-1)
      /time_rec_interval*time_rec_interval);  // 12 Apr 10, was +1

   n_flow_watchers = flow_watcher_control_space.mx_objects;
   n_fw_running = 0;
   fw_next_st_to_check = n_flow_watchers;
   reset_summary_info();
   last_drops = 0;

#if STARTUP_DEBUG
   printf("about to start flow-watch threads +-+\n");  fflush(stdout);
#endif
   /* Start flow-watching thread(s) */
   ht_size = stream_space.mx_objects*(MX_EXTRA_BLOCKS+1)*2/mx_flow_watchers;
   if (ht_size < MIN_HT_SIZE) ht_size = MIN_HT_SIZE;
#if STARTUP_DEBUG
      printf("start_meter(2.5):mx_stream=%u, mx_fw=%u, ht_size=%u\n",
         stream_space.mx_objects,mx_flow_watchers, ht_size);  //d
#endif
   for (j = 0; j != flow_watcher_control_space.mx_objects; ++j)
      start_flow_watcher(&flow_watcher_control_space.objects[j],
         pw_q_sz, ht_size);

#if STARTUP_DEBUG
   printf("about to start packet-wat threads +-+\n");  fflush(stdout);
#endif
   /* Start the packet-watching thread(s) */
   for (n_interfaces_running = j = 0; j != n_interfaces; ++j) {
      iip = &if_info[j];
      if (pthread_create(&iip->pw_thread, NULL, packet_watcher_process, iip)) {
         fprintf(stderr, "Error creating packet-watch thread <<<\n");
         exit(31);
         }
      iip->thread_running = 1;  n_interfaces_running += 1;
      }

   setgid(ut_gid);  setuid(ut_uid);  /* Don't need to run as root now */
   my_uid = ut_uid;  my_gid = ut_gid;

   log_msg(LOG_WARNING, 0, "Starting %s, running on %s",
      version, node.nodename);

#if STARTUP_DEBUG
   printf("leaving start_meter() +-+\n");  fflush(stdout);
#endif
   }

void stop_meter(char const *msg) {
   struct stream_txt *st;
   st = get_stream_txt_object(&stream_txt_space);
   strcpy(st->msg, msg);
   enqueue_stream_txt(st->msg);  /* Queue shutdown message */
   }

int check_stream_chains(char const *msg, struct stream_queue *sq)
{  /* Consistency check for stream queue */
   struct stream *fsp,*rsp, *l_fsp, *l_rsp;
   int k, nas, rv;

   nas = sq->length;
   fsp = sq->head;  rsp = sq->tail;
   l_fsp = l_rsp = NULL;

   for (rv = 1, k = 0; ; ++k) {  /* Walk forward and reverse chains */
      if (fsp == NULL || rsp == NULL) {  /* Both NULL when done */
         if (fsp != NULL || fsp != rsp) {
	    log_msg(LOG_WARNING, 0,
               "s_c_c(%s): fsp=%x != rsp=%x, k=%d. nas=%d\n",
		    msg, fsp, rsp, k, nas);
 quack(419);
            rv = 0;
            }
         break;
         }
      if (k > nas+10) {
         log_msg(LOG_WARNING, 0,
            "s_c_c(%s): loop in chain, k=%d, nas=%d", msg, k, nas);
         fflush(stdout);
 quack(420);
         rv = 0;
         break;
         }
      l_fsp = fsp;  l_rsp = rsp;  /* For gdb debugging */
      fsp = fsp->next;  rsp = rsp->prev;
      }

   if (k != nas) {
      log_msg(LOG_WARNING, 0,
         "s_c_c(%s): k=%u != nas=%u", msg, k, nas);
 quack(421);
      printf("\n");
      rv = 0;
      }
   return rv;
   }

double at_seconds(struct timeval tv) {  /* Arrival time in seconds */
   return (double)(tv.tv_sec-start_ts) + (double)tv.tv_usec/1000000.0;
   }

uint32_t pkt_drops(void) {
   uint32_t j;  uint64_t r, drops = 0;
   for (j = 0; j != n_interfaces; ++j)
      drops += trace_get_dropped_packets(if_info[j].trace);
   r = drops < last_drops ? 0 : drops - last_drops;
   last_drops = drops;
   return (uint32_t)r;
   }

int init_trace(struct if_info *iip) {
   struct timeval tv;

   iip->trace = NULL;  iip->packet = NULL;
   iip->trace = trace_create(iip->if_name);
   if (trace_is_err(iip->trace)) {
      log_msg(LOG_ERR, 0, "Couldn't open trace %s <<<", iip->if_name);
      return 101;
      }

   snaplen = SNAP_LEN;
   if (trace_config(iip->trace,TRACE_OPTION_SNAPLEN,&snaplen)) {
      log_msg(LOG_ERR, 0, "Couldn't set snap length %d <<<", snaplen);
      return 102;
      }

   if (trace_start(iip->trace)) {
      log_msg(LOG_ERR, 0, "Couldn't start trace %s <<<", iip->if_name);
      trace_destroy(iip->trace);
      return 103;
      }

   iip->packet = trace_create_packet();
   if (trace_read_packet(iip->trace, iip->packet) > 0) {
      if (start_ts == 0) {
         tv = trace_get_timeval(iip->packet); 
         start_ts = (uint32_t)tv.tv_sec;
         set_date_seed(start_ts);  /* For Conficker p2p test */
         }
      }
   else {
      log_msg(LOG_ERR, 0, "Couldn't read packet from %s <<<", iip->if_name);
      return 104;
      }

   return 0;
   }

void close_trace(struct if_info *iip) {
   if (iip->trace != NULL) {
      if (trace_is_err(iip->trace)) log_msg(LOG_ERR, 1,
         "Error reading packets from %s !!!", iip->if_name);
      trace_destroy(iip->trace);
      }
   if (iip->packet != NULL)
      trace_destroy_packet(iip->packet);
   }

int next_trace_file(struct if_info *iip) {
   free(iip->if_name);  /* processewd uri */
   iip->if_name = other_uris[next_uri_nbr];
   next_uri_nbr += 1;

   iip->trace = NULL;  iip->packet = NULL;
   iip->trace = trace_create(iip->if_name);
   if (trace_is_err(iip->trace)) {
      log_msg(LOG_ERR, 0, "Couldn't open trace %s <<<", iip->if_name);
      return 101;
      }

   if (trace_config(iip->trace,TRACE_OPTION_SNAPLEN,&snaplen)) {
      log_msg(LOG_ERR, 0, "Couldn't set snap length %d <<<", snaplen);
      return 102;
      }

   if (trace_start(iip->trace)) {
      log_msg(LOG_ERR, 0, "Couldn't start trace %s <<<", iip->if_name);
      trace_destroy(iip->trace);
      return 103;
      }

   iip->packet = trace_create_packet();
   return 0;
   }

void one_second_process(uint32_t ts) {
   struct if_info *iip;
   uint32_t j, active_opq_entries, active_streams, 
      active_v4_flow_infos, active_v6_flow_infos;

   last_ts = ts;  /* Save ts for end-of-run */
   set_date_seed(ts);  /* For Conficker p2p test */

   for (j = 0; j != n_interfaces; ++j) {
      iip = &if_info[j];
      iip->av_pkt_count += iip->n_pkts;  iip->n_seconds += 1;
      if (iip->n_pkts > iip->mx_pkt_rate) iip->mx_pkt_rate = iip->n_pkts;
      iip->n_pkts = 0;  g_last_ts = iip->last_ts = ts;
      }
   mutex_unlock("osp", &if_info_mutex);  /* Locked in seconds_change() */

   active_opq_entries = opq_entry_space.mx_objects -
      opq_entry_space.free_queue.length;
   if (active_opq_entries > mx_active_opq_entries)
      mx_active_opq_entries = active_opq_entries;

   active_streams = stream_space.mx_objects -
      stream_space.free_queue.length;
   if (active_streams > mx_active_streams)
      mx_active_streams = active_streams;

   active_v4_flow_infos = v4_flow_info_space.mx_objects -
      v4_flow_info_space.free_queue.length;
   if (active_v4_flow_infos > mx_active_v4_flow_infos)
      mx_active_v4_flow_infos = active_v4_flow_infos;
   active_v6_flow_infos = v6_flow_info_space.mx_objects -
      v6_flow_info_space.free_queue.length;
   if (active_v6_flow_infos > mx_active_v6_flow_infos)
      mx_active_v6_flow_infos = active_v6_flow_infos;
   }

#if STOP_AFTER
int n_times = STOP_AFTER;
#endif
void seconds_change(struct if_info *iip, uint32_t ts) {
   uint32_t e_type;
   struct stream *event_sp;
   struct flow_watcher_control *fw;

   mutex_lock("osp", &if_info_mutex);
   if (ts <= g_last_ts) {
      iip->last_ts = ts;
      mutex_unlock("osp", &if_info_mutex);
      }
   else {  /* Change in overall last_ts */
      one_second_process(ts);  /* Update g_last_ts, unlock if_info */

      for (;;) {  /* Handle events */
         e_type = check_events(ts);
         if (e_type == NO_EVENT) break;
	 else switch (e_type) {
	 case S_TIME_EVENT:  /* Timeout streams */
	    if ((event_sp = get_stream(6)) == NULL)
	       break;  /* Can't schedule another TIME_EVENT */

   	    fw_next_st_to_check -= 1;
	    fw = &flow_watcher_control_space.objects[
               fw_next_st_to_check];
            if (fw_next_st_to_check == 0)
               fw_next_st_to_check = n_flow_watchers;

	    t_semaphore_down(&fw->ispq_empty);
            event_sp->event_type = S_TIME_EVENT;
            event_sp->ts = g_last_ts;

            stream_q_lock(&fw->ispq);
            stream_push(&fw->ispq, event_sp);
            stream_q_unlock(&fw->ispq);
            t_semaphore_up(&fw->ispq_ready);
               /* Pass event to flow watcher */
            break;
	 case STATS_EVENT:  /* Write #Stats */
	    update_interface_stats(ts);
#if STOP_AFTER
            if (--n_times == 0) request_shutdown = 1;
#endif
            break;
         case TIME_EVENT:  /* Write #Time */
	    each_time_rec(ts);
            break;
            }
         }  /* event handling loop */
      }
   }

int n_tcpks = 0;    //tcpk

void each_pkt(struct flow_watcher_control *fw, struct stream *isp) {
   struct stream *slr, *tq;
   int new = 1;

   isp->hash = owt_hash(isp);
   slr = owt_ht_lookup(fw->sh_table, isp->hash, isp, &new);
                                         /* new => OK_to_insert */
   if (new) {  /* New stream, slr == isp */
      slr->state = SS_ONE_WAY;
      slr->version = isp->version;
      slr->first_fwd = isp->fwd;
      slr->tb = slr->fb = slr->tp = slr->fp = slr->totp = 0;
      slr->first_us = slr->last_us;
      slr->last_report_sec = slr->ts;
      slr->timeout = STO_ONE_WAY;
      slr->src_addr = isp->src_addr;
      slr->inactive_last_sum = 0;
      slr->ts = isp->ts;
      slr->tcp_key = isp->tcp_key;  /* Only want the SYN key */
      get_flow_info(slr);
#if ERF_DIRECTION
      slr->direction = isp->direction;
#endif
      //      if (slr->proto == 6) fprintf(tcp_keys_f, "%8x\n", slr->tcp_key);  //tcpk
      //      if ((n_tcpks += 1) == 10000) { fclose(tcp_keys_f);  exit(123); }  //tcpk
      }
   else {  /* Already in table, copy fields from isp */
      if (slr == isp) quack(3004);
      slr->last_us = isp->last_us;
#if ERF_DIRECTION
      slr->direction |= isp->direction;  /* Keep track of direction for *all* pkts */
#endif
      }

   if (isp->fwd) {  /* !local -> local, i.e. pkts in to home network */
      mutex_lock("epk", &stream_data_mutex);
      slr->tp += 1;  slr->tb += isp->ip_len;
      mutex_unlock("epk", &stream_data_mutex);
      }
   else {
      mutex_lock("epk", &stream_data_mutex);
      slr->fp += 1;  slr->fb += isp->ip_len;
      mutex_unlock("epk", &stream_data_mutex);
      }
   slr->totp += 1;
   slr->active_since_last_sum = 1;

   if (slr->state != SS_TWO_WAY) {  /* Not yet SS_TWO_WAY */
      if (slr->fp != 0) {  /* Pkts out from home network -> two-way stream */
         slr->state = SS_TWO_WAY;
         slr->timeout = STO_TWO_WAY;
 	 free_flow_info(slr);  /* Don't need flow info now */
         }
      else {  /* Still one-way, update flow info */
	 if (slr->flow_info.fi4) {  /* i.e. flow_info != NULL */
            int half_flow = update_flow_info(slr, isp);
#if HALF_WAY_STREAMS
            if (slr->state != SS_HALF_WAY && half_flow) {
               slr->state = SS_HALF_WAY;
               slr->timeout = STO_ONE_WAY;
 	       free_flow_info(slr);  /* Don't need flow info now */
	       }
#endif
	    }
         else {
   	    slr->ts = isp->ts;
            isp->cfc = 0;  /* SRI test for Conficker p2p source */
            /* Any UDP, any TCP not to port 445 or from port 80 */
            if (slr->version == 4 && (
                  isp->proto == PT_UDP || (isp->proto == PT_TCP &&
	          !(isp->dst_port == 445 || isp->src_port == 80)) )) {
	       int rc, i;	 result_t res;
 	       rc = portgen(isp->dst_addr.a.v4, &res, date_seed);
	       for (i = 0; i < 8; i++) {
	          if (res.u16[i] == isp->dst_port) {
	             isp->cfc = 1;
	             break;  /* Need only match one port! */
	             }
	          }
               }
	    }
         }
      }
#if HALF_WAY_STREAMS
   else if (slr->state == SS_HALF_WAY) {
      }
#endif
   else if (slr->state == SS_DELETE) {
      slr->state = SS_ONE_WAY;  /* Was inactive for a Record interval */
      }
   else if (slr->state == SS_TWO_WAY) {
      }
   else log_msg(LOG_ERR, 0, "stream in state %d <<<<", slr->state);

   if (new) {
      stream_q_lock(&fw->sq);
      stream_push(&fw->sq, slr);
      stream_q_unlock(&fw->sq);
      }
   else {
      free_stream(isp);  /* Finished with isp */
      stream_q_lock(&fw->sq);
      tq = stream_extract(&fw->sq, slr);  /* Move slr to stream list tail */
         /* check_stream_timeouts() starts from head,
	    so walk_up() starts from the most recently active */
      stream_push(&fw->sq, tq);
      stream_q_unlock(&fw->sq);
      }
#if SC_DEBUG >= 3
   check_stream_chains("+++ end of each_pkt", &fw->sq);
#endif
   }

void shutdown_flow_watchers(void) {
   struct flow_watcher_control *fw;
   int k;
   for (k = 0; k != n_flow_watchers; ++k) {
     printf("shutdown_flow_watchers: k=%d\n", k);  // XXX
      fw = &flow_watcher_control_space.objects[k];
      t_semaphore_down(&fw->ispq_empty);  /* Wait for an empty entry */
      fw->eof_sp->event_type = END_OF_FILE;
      stream_q_lock(&fw->ispq);
      stream_push(&fw->ispq, fw->eof_sp);
      stream_q_unlock(&fw->ispq);
      t_semaphore_up(&fw->ispq_ready);  /* Signal a full entry */
      }
   }

void cancel_packet_watchers(void) {
   uint32_t j;  int r;
   for (j = 0; j != n_interfaces; ++j) {
      if (if_info[j].thread_running) {
         r = pthread_cancel(if_info[j].pw_thread);
         printf("thread %d cancel -> %d\n", j,r);  fflush(stdout);
         }
      }
   shutdown_flow_watchers();
   }

uint32_t nfw[4] = {0, 0, 0, 0};

void *packet_watcher_process(void *arg) {
   struct if_info *iip = (struct if_info *)arg;
   struct timeval tv;
   uint16_t ethertype;   uint32_t hl;  libtrace_ip_t *ip;
   int key_code, fwd;
   struct stream *new_sp;
   uint32_t ts, hash;
   struct flow_watcher_control *fw;
   uint32_t fn = 0;
   uint8_t *sap;  // XXX

   //##   fwd = nice(-5);  /* Run this thread at high priority */
      /* Result not used, keep compiler happy */

   iip->last_ts = start_ts;

   for (;;) {  /* Loop through files in list */
      while (trace_read_packet(iip->trace, iip->packet) > 0) {
         if (request_shutdown)  /* User sent SIGINT */
       	    break;  /* Test here; pthread_cancel seems to clobber iip->trace */
         iip->n_pkts += 1;  /* For av/max pkts #stats */
         ip = trace_get_layer3(iip->packet, &ethertype, &hl);
         if (!ip || (ethertype != 0x0800 && ethertype != 0x86dd))
	    n_not_ip += 1;
         else {
            tv = trace_get_timeval(iip->packet);
            ts = (uint32_t)tv.tv_sec;
            if (ts > iip->last_ts) seconds_change(iip, ts);

            key_code = ethertype == 0x0800 ?
	       get_v4key(&new_sp, &fwd,  hl, (uint8_t *)ip)
 	       : get_v6key(&new_sp, &fwd,  hl, (uint8_t *)ip);
            if (key_code != SKEY_OK) {
               /* Couldn't get packet key (packet freed by get_key) */
               switch (key_code) {
               case SKEY_NOPEERADDR:
               case SKEY_NOTRANSADDR:
	          n_too_small += 1;
                  break;
               case SKEY_FRAGPKT:
                  n_frags += 1;
                  break;
               case SKEY_NOLOCALADDR:
                  n_no_local_address += 1;
                  break;
               case SKEY_V6OPAQUE:
                  n_v6_opaque += 1;
                  break;
	       case SKEY_NOSPACE:
	 	  break;  /* Ignore, get_stream displayed gthe message */
               default:
                  log_msg(LOG_ERR, 66,
	             "packet_watcher: bad key_code %d !!!", key_code);
                  break;
                  }
	       }
            else {
               if (n_flow_watchers == 1) new_sp->fw_nbr = 0;
	       else {
                  sap = new_sp->src_addr.ver == 4 ?
		     &new_sp->src_addr.a.a[IP4_ADDR_LEN-1] :
		     &new_sp->src_addr.a.a[IP6_ADDR_LEN-1];
                  new_sp->fw_nbr = (sap[-1]) % n_flow_watchers;
	          }
#if 0
	      new_sp->fw_nbr = (n_flow_watchers == 1 ? 0 :
                  (new_sp->src_addr.ver == 4 ?
		  new_sp->src_addr.a.a[IP4_ADDR_LEN-1] :
		  new_sp->src_addr.a.a[IP6_ADDR_LEN-1]) )
		  % n_flow_watchers;

               new_sp->fw_nbr = n_flow_watchers == 1 ? 0 :
		  // (hash >> 24) % n_flow_watchers;
                  hash % n_flow_watchers;
#endif
	       nfw[new_sp->fw_nbr] += 1;
               fw = &flow_watcher_control_space.objects[new_sp->fw_nbr];

               t_semaphore_down(&fw->ispq_empty);  /* Wait for empty entry */
               new_sp->event_type = NO_EVENT;
	       // XXXXX               new_sp->hash = hash;
               new_sp->fwd = fwd;
               new_sp->wl = trace_get_wire_length(iip->packet);
               new_sp->ts = ts;
               new_sp->last_us = at_seconds(tv);
#if ERF_DIRECTION
 	       new_sp->direction = trace_get_direction(iip->packet)+1;
	          /* 0 => portA, 1 => portB */
#endif
               stream_q_lock(&fw->ispq);
               stream_push(&fw->ispq, new_sp);
               stream_q_unlock(&fw->ispq);
               t_semaphore_up(&fw->ispq_ready);
                  /* Pass pkt_info to selected flow watcher */
   	       }
            }
         if (request_shutdown) break;  /* User sent SIGINT */
         }  /* SIGINT or trace file EOF */

      if (request_shutdown) break;  /* User sent SIGINT */
      fn += 1;
      if (fn == traces_to_process) break;  /* No more files */

      close_trace(iip);
      if (next_trace_file(iip) != 0)
         log_msg(LOG_ERR, 1, "Couldn't start next file !!!");
      log_msg(LOG_WARNING, 0, "----- opened %s", iip->if_name);
      }

   iip->thread_running = 0;
   n_interfaces_running -= 1;
   printf("EOF, n_interfaces_ruunning=%d\n", n_interfaces_running);  // XXX
   if (n_interfaces_running == 0)
      shutdown_flow_watchers();
   return NULL;  /* Keep compiler happy */
   }

void *flow_watcher_process(void *arg) {
   struct flow_watcher_control *fw = arg;
   struct stream *isp;
   uint32_t shutdown_now, rake_reqd, j;

   //##   j = nice(-3);  /* Run this thread at medium priority */
      /* Result not used, keep compiler happy */
#if STARTUP_DEBUG
   printf("starting flow_watcher_process() +-+\n"); fflush(stdout); 
#endif
   for (;;) {
      t_semaphore_down(&fw->ispq_ready);  /* Wait for a full block */
      stream_q_lock(&fw->ispq);
      isp = stream_shift(&fw->ispq);
      stream_q_unlock(&fw->ispq);

      if (isp->event_type == END_OF_FILE) {
	printf("flow_watcher(%d): EOF\n", fw->id);  // XXX
         break;
         }
      else switch (isp->event_type) {
      case S_TIME_EVENT:
	printf("watcher(%d): TIME_EVENT\n", fw->id);  // XXX
        rake_reqd = n_extra_streams == MX_EXTRA_BLOCKS &&  
	   stream_space.free_queue.length < MN_FREE_STREAMS;
	if (STREAMTIMEOUTS || rake_reqd) {
            mutex_lock("recording", &count_mutex);
	    if (CV_RECORDING)  /* Handling #Record: don't try to rake */
	       mutex_unlock("recording", &count_mutex);
	    else {
  	       count += 1;  /* About to run check_stream_timeouts() */
               mutex_unlock("recording", &count_mutex);

	       if (rake_reqd)
                  j = rake_idle_streams(fw, isp->ts);
               else if (STREAMTIMEOUTS) 
                  check_stream_timeouts(fw, isp->ts);

    	       pthread_mutex_lock(&count_mutex);
	       count -= 1;  /* check_stream_timeouts() finished */
 	       pthread_cond_signal(&count_cv);  /* Let Record: know! */
	       pthread_mutex_unlock(&count_mutex);
	       }
	    }
	 isp->event_type = 0;  free_stream(isp);
	 break;
      default:
	 each_pkt(fw, isp);
	 break;
         }

      t_semaphore_up(&fw->ispq_empty);  /* Signal an empty block */
      }  /* Loop finishes on interrupt or end of trace file */

      //xx  output_active_streams(&fw->sq, isp->ts);
      /* Dump currently active streams */

   mutex_lock("fwp", &n_fw_mutex);
   n_fw_running -= 1;
   shutdown_now = n_fw_running == 0;
   mutex_unlock("fwp", &n_fw_mutex);

   if (shutdown_now) {
      for (j = 0; j != n_interfaces; ++j)
         close_trace(&if_info[j]);
      if (request_shutdown) 
         stop_meter("#Shutdown: User sent interrupt ###");
      else {
	 printf("nfw[] = %u, %u, %u, %u\n", nfw[0], nfw[1], nfw[2], nfw[3]);
         stop_meter("#Shutdown: End of trace file ###");
         }
      }
   return NULL;  /* Keep compiler happy */
   }
