#!/usr/local/bin/ruby

# 1410, Fri 22 Apr 11 (NZST)
#
# monitorinfo.rb: Parses owt-config.rb files, yielding a set
#                   of Ruby symbols in @config
#
# iatmon/owtmon: One-Way Traffic Monitor testbed
# Copyright (C) 2009-2011 by Nevil Brownlee, U Auckland | CAIDA | WAND
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>


class MonitorInfo
  def next_token
    loop do    
      while @tokens.size == 0
        begin 
          @tokens = @f.readline.chomp.split(' ')
        rescue EOFError
          return nil
        end
      end
      tok = @tokens.shift
      if tok[0].chr == '#'
        @tokens = Array.new
      else
        # print "tok=>#{tok}<\n"
        return tok
      end
    end
  end

  def initialize(f)
    @f = f
    @config = Hash.new
    @tokens = Array.new  # Parse ???-owt-config.rb file
    loop do
      var = next_token
      break if var.nil?
      equal_sign = next_token
      print "Expected equal sign (=) !!\n" if equal_sign != '='
      val = next_token
      if val[0].chr == "'"
        t = val
        while t[-1].chr != "'"
          t = next_token;  val += " #{t}"
        end
        val = val[1..-2]  # Remove ''
      end
      # print ">#{var}< #{equal_sign} >#{val}\n"
      @config[:"#{var}"] = val
    end
    @f.close

    v6_in_list = false  # Create derived config info
    @config[:LocalAddress].split(/,\s*/).each do |la|
      v6_in_list = true if la =~ /:/
    end
    @config[:v6_in_list] = v6_in_list
    # print ":v6_in_list=#{@config[:v6_in_list]}\n" if v6_in_list

    meter_loc = @config[:MeterLoc].downcase
    @config['meter_loc'] = meter_loc
    sum_dir = @config[:SummaryDir]
  end

  def to_s
    @config.each { |key, val|  print "  #{key} => #{val}\n"}
    print "\n"
  end

  attr_reader :config
end  # MonitorInfo
