#!/usr/bin/env ruby

#  1627, Wed 22 Dec 10 (NZDT)
#  0955, Fri 26 Feb 10 (NZDT)
#  1818, Sun  9 Aug 09 (NZST)
#
#  lib/owt_monitor.rb: Class Streamer, user interface to OWTmonitor
#
# iatmon/owtmon: One-Way Traffic Monitor testbed
# Copyright (C) 2009-2011 by Nevil Brownlee, U Auckland | CAIDA | WAND
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

require 'optparse'

def streamer_usage()
  $stderr.print <<END
Usage: #{File.basename $0} [ -i interface ] [ -c interval ] [-l location]
         [ -d pktdatas ] [ -s streams ] [ -n minpkts ] [-o logname] [ -t types ] 
         [ -u ]  # to build source distributions
END
end

module OWTmonitor
  class Monitor
    def usage(status, msg = nil)
      $stderr.puts msg if msg
      streamer_usage
      exit(status)
    end

    attr_reader :live, :n_interfaces, :last_fn

    def check_interfaces(interfaces)
      @live = nil
      interfaces.each do |if_name|
        if if_name =~ /^(.+):(.+)$/
          scheme = $1;  @last_fn = $2
          live = !(scheme == 'pcap' ||scheme == 'pcapfile' || \
            scheme == 'erf')
          $stderr.print \
            "interface=#{if_name}, scheme=#{scheme}, live=#{live}\n"
          if @live.nil?
            @live = live
          elsif live != @live
            print "Interfaces must be all live or all !live <<<\n"
            exit
          end
        else
          $stderr.print "No : in #{if_name} (libtrace expects a URI) <<<\n"
          exit
        end
      end
      @n_interfaces = interfaces.size
    end

    def init(args = nil)
      print "args=>#{args}<\n"

      interfaces = Array.new
      streams = 90000;  min_pkts = 10;  meter_loc = ''
      timerecinterval = 300;  types = 10;  hours = 1
      logname = 'owt.log'
      distributions = 0  # Default = no distributions
      if args
        ARGV[0,0] = args.split(/\s+/)
           # Prepend to ARGV, so that command-line args 
           # will follow whatever we're given as a parameter
      end
      opts = OptionParser.new do |opts|
        opts.on('-c INTERVAL') { |v| timerecinterval = v.to_i }
        opts.on('-h HOURS') { |v| hours = v.to_i }
        opts.on('-i INTERFACES') { |v| interfaces.push(v) }
        opts.on('-l LOCATION') { |v| meter_loc = v }
        opts.on('-n MINPKTS') { |v| min_pkts = v.to_i }
        opts.on('-o LOGNAME') { |v| logname = v }
        opts.on('-s STREAMS') { |v| streams = v.to_i }
        opts.on('-t TYPES') { |v| types = v.to_i }
        opts.on('-u DISTRIBUTIONS') { |v| distributions = v == 'D' ? 2 : 1}
      end
      opts.parse!(ARGV) 
#       print "interface=#{interfaces}" \
#         + ", streams=#{streams}, timeinterval=#{timerecinterval}\n"
      if interfaces.size == 0
        $stderr.print "No interface(s) specified <<<\n"
        exit
      end
      check_interfaces(interfaces)
      self.owt_init(interfaces, @live, streams,
        timerecinterval, min_pkts, types, hours, meter_loc, logname,
        distributions)
    end  # Monitor.init

  # Only the usage and init methods (above) are in Ruby - 
  # Monitor's other methods are in owt_monitor.c.  They are ..
  #
  #  initialize  - creates new Monitor object, calls owt_init
  #  owt_init    - does C-level intialising for init
  #  ...

  end  # class Monitor
end

Monitor = OWTmonitor::Monitor  # Allow us to use Monitor.new directly
