#!/usr/bin/env ruby

#  2212, Tue  8 Jun 10 (NZST)
#
#  write-flow-sum(): writes a summary file
#
# iatmon/owtmon: One-Way Traffic Monitor testbed
# Copyright (C) 2009-2011 by Nevil Brownlee, U Auckland | CAIDA | WAND
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>


def write_sum_file(fname, start_ts, now_ts)
  $osf = File.new(fname, 'w')
  $osf.print "#version\n"
  $osf.print "#{FormatVer} #{OWTmonVer} #{Meter_ver}\n\n"

  $osf.print "#meter_info\n"
  $osf.print "#{MeterLoc} #{utc_gts(start_ts)}" +
    " #{Trace_sz} #{EnoughPkts} #{MinPkts}\n\n"

  src_type_counts = Array.new(N_source_types, 0)
  tcp_type_counts = Array.new(N_source_types, 0)
  udp_type_counts = Array.new(N_source_types, 0)
  other_type_counts = Array.new(N_source_types, 0)
  tcp_probe_port_counts = NbrHash.new;  udp_probe_port_counts = NbrHash.new
#  all_probe_port_counts = NbrHash.new
  lifetime_distribs = Array.new;  pkt_rate_distribs = Array.new
  etot_ratio_distribs = Array.new
  retry_time_distribs = Array.new;  dport_distribs = Array.new
  pkt_len_distribs = Array.new
  pkt_count_distribs = Array.new
  src_count_distribs = Array.new
  tot_count_distribs = Array.new
  (0..N_source_types-1).each do
    lifetime_distribs.push(Distribution.new(1.0, MxLifetime, 120, DT_LOG, 1.0))
    pkt_rate_distribs.push(Distribution.new(0.001, 1000.0, 100, DT_LOG, 1.0))
    etot_ratio_distribs.push(Distribution.new(0.010, 100.0, 100, DT_LOG, 1.0))
    retry_time_distribs.push(Distribution.new(0.12, 120.0, 100, DT_LOG, 1.0))
    dport_distribs.push(Distribution.new(0.4, EnoughPkts+0.4, 1+EnoughPkts,
      DT_LINEAR, 1.0))
    pkt_len_distribs.push(Distribution.new(40, 1500, 100, DT_LOG, 1.0))
    pkt_count_distribs.push(Distribution.new(1, 5000, 100, DT_LOG, 1.0))
    src_count_distribs.push(Distribution.new(0, 255, 256, DT_LINEAR, 1.0))
    if TwoWay
      tot_count_distribs.push(Distribution.new(1, 50000, 120, DT_LOG, 1.0))
    end
  end

  $owt.each_stream do |s|  # look at final attribs for all streams
    src_type, probe_port, enough_lifetime, to_pkts, total_lifetime,
      mean_retry_time, n_dest_ports, pkt_len_mode, src_bin,
      tot_count, tcp_pkts, udp_pkts, other_pkts  = s
    src_type_counts[src_type] += 1
    tcp_type_counts[src_type] += tcp_pkts
    udp_type_counts[src_type] += udp_pkts
    other_type_counts[src_type] += other_pkts
    if !probe_port.nil?
      tcp_probe_port_counts.update(probe_port) if src_type == T_tcp_probe
      udp_probe_port_counts.update(probe_port) if src_type == T_udp_probe
#      all_probe_port_counts.update(probe_port)
    end

    lifetime_distribs[src_type].count(total_lifetime)
    total_rate = to_pkts < MinPkts ? 0 : (to_pkts - 1)/total_lifetime
    pkt_rate_distribs[src_type].count(total_rate)
# if total_rate > 100
#   printf("rate=%.1f, to_pkts=%d, lifetime=%.2f, src_type%d\n",
#     total_rate, to_pkts, total_lifetime, src_type)
# end
    if to_pkts >= EnoughPkts  # Enough (for total_lifetime > enough_lifetime)
      enough_rate = (EnoughPkts - 1)/enough_lifetime
      etot_ratio_distribs[src_type].count(enough_rate/total_rate)
    end
    pkt_count_distribs[src_type].count(to_pkts)
    if !mean_retry_time.nil?
      retry_time_distribs[src_type].count(mean_retry_time)
      dport_distribs[src_type].count(n_dest_ports)
      pkt_len_distribs[src_type].count(pkt_len_mode)
    end
    src_count_distribs[src_type].count(src_bin) if !src_bin.nil?
    if TwoWay
      tot_count_distribs[src_type].count(tot_count)
    end
  end
  way_counts, st_distribs = $owt.c_addr_distribs

  $osf.print "#source_types\n"
#?  sc = $owt.src_counts
#?  $osf.print "#{sc.size}"
#?  sc.each { |c| $osf.print " #{c}" }
  $osf.print N_source_types
  src_type_counts.each { |c| $osf.print " #{c}" };  $osf.print "\nTCP "
  tcp_type_counts.each { |c| $osf.print " #{c}" };  $osf.print "\nUDP " 
  udp_type_counts.each { |c| $osf.print " #{c}" };  $osf.print "\nOTHER "
  other_type_counts.each { |c| $osf.print " #{c}" };  $osf.print "\n\n"

#  $osf.print "#probe_ports\n#{N_ProbePorts}"
#  all_probe_port_counts.write($osf)
#  $osf.print "\n\n"

  $osf.print "#tprobe_ports\n#{N_ProbePorts}"
  tcp_probe_port_counts.write($osf)
  $osf.print "\n\n"

  $osf.print "#uprobe_ports\n#{N_ProbePorts}"
  udp_probe_port_counts.write($osf)
  $osf.print "\n\n"

  $osf.print "#lifetimes\ns "
  lifetime_distribs[0].write_x($osf, 4)
  lifetime_distribs.each_index do |st|
    $osf.print "#{st} "
    lifetime_distribs[st].write_y($osf, false)
  end
  $osf.print "\n"

  $osf.print "#pktrates\np/s "
  pkt_rate_distribs[0].write_x($osf, 4)
  pkt_rate_distribs.each_index do |st|
    $osf.print "#{st} "
    pkt_rate_distribs[st].write_y($osf, false)
  end
  $osf.print "\n"

  $osf.print "#etotratios\nf "
  etot_ratio_distribs[0].write_x($osf, 4)
  etot_ratio_distribs.each_index do |st|
    $osf.print "#{st} "
    etot_ratio_distribs[st].write_y($osf, false)
  end
  $osf.print "\n"

  $osf.print "#retrytimes\ns "
  retry_time_distribs[0].write_x($osf, 4)
  retry_time_distribs.each_index do |st|
    $osf.print "#{st} "
    retry_time_distribs[st].write_y($osf, false)
  end
  $osf.print "\n"

  $osf.print "#dports\ni "
  dport_distribs[0].write_x($osf, 1)
  dport_distribs.each_index do |st|
    $osf.print "#{st} "
    dport_distribs[st].write_y($osf, false)
  end
  $osf.print "\n"

  $osf.print "#pktlengths\nB "
  pkt_len_distribs[0].write_x($osf, 1)
  pkt_len_distribs.each_index do |st|
    $osf.print "#{st} "
    pkt_len_distribs[st].write_y($osf, false)
  end
  $osf.print "\n"

  $osf.print "#pktcounts (this hour)\ni "
  pkt_count_distribs[0].write_x($osf, 1)
  pkt_count_distribs.each_index do |st|
    $osf.print "#{st} "
    pkt_count_distribs[st].write_y($osf, false)
  end
  $osf.print "\n"

  if !st_distribs.nil?
    dest_addresses = st_distribs  # Only one so far

    $osf.print "#dstaddresses\ni "  # Dummy up x-values line
    dest_addresses[0].each_index { |st| $osf.print " #{st}" }
    $osf.print "\n"
    dest_addresses.each_index do |st|
      $osf.print "#{st} "
      da_counts = dest_addresses[st]
      da_counts.each { |c| $osf.print " #{c}" }
      $osf.print "\n"
    end
    $osf.print "\n" 
 
    $osf.print "#srcaddresses\ni "
    src_count_distribs[0].write_x($osf, 1)
    src_count_distribs.each_index do |st|
      $osf.print "#{st} "
      src_count_distribs[st].write_y($osf, false)
    end
    $osf.print "\n"
  end

  ist = $owt.interval_stats
  if !ist.nil?
    #x  active_sources, unique_hosts, pkts_seen = ist
    active_sources, unique_hosts = ist

    $osf.print "#active_sources\n"
    active_sources.each_index do |st|
      $osf.print "#{st} "
      as = active_sources[st]
      as.each { |c| $osf.print " #{c}" }
      $osf.print "\n"
    end

    $osf.print "\n"
    $osf.print "#unique_hosts\n"
    unique_hosts.each_index do |st|
      $osf.print "#{st} "
      uh = unique_hosts[st]
      uh.each { |c| $osf.print " #{c}" }
      $osf.print "\n"
    end
    $osf.print "\n"

  #x  $osf.print "#pkts_seen\n"
  #x  pkts_seen.each_index do |st|
  #x    $osf.print "#{st} "
  #x    ps = pkts_seen[st]
  #x    ps.each { |c| $osf.print " #{c}" }
  #x    $osf.print "\n"
  #s  end
  #x  $osf.print "\n"
  end

  if TwoWay
    $osf.print "\n"
    $osf.print "#totcounts (source lifetime)\nB  "
    tot_count_distribs[0].write_x($osf, 1)
    tot_count_distribs.each_index do |st|
      $osf.print "#{st} "
      tot_count_distribs[st].write_y($osf, false)
    end
    $osf.print "\n"

    $osf.print "#waycounts\n"
    $osf.print "#{way_counts[0]}, #{way_counts[1]}"    # 1-way pkts, bytes
    $osf.print "  #{way_counts[2]}, #{way_counts[3]}"  # 2-way pkts, bytes
    $osf.print "  #{way_counts[4]}, #{way_counts[5]}"  # 1 and 2-way streams
    $osf.print "  #{way_counts[6]}\n"  # not-one-way-streams
    $osf.print "#{way_counts[7]}, #{way_counts[8]}"    # v6 1-way pkts, bytes
    $osf.print "  #{way_counts[9]}, #{way_counts[10]}"  # v6 2-way pkts, bytes
    $osf.print "  #{way_counts[11]}, #{way_counts[12]}\n"  # v6 1 & 2-way streams
    $osf.print "\n"
  end

  fin_proc_time = Process::times;  fin_proc_date = Time.now
  time_string = sprintf("(minutes): trace_duration=%.2f" +
      ", elapsed=%.3f, user=%.3f, system=%.3f",
    (now_ts - start_ts)/60.0,
    (fin_proc_date.to_i - $start_proc_date.to_i)/60.0,
    (fin_proc_time.utime - $start_proc_time.utime)/60.0,
    (fin_proc_time.stime - $start_proc_time.stime)/60.0 )
  $start_proc_time = Process::times;  $start_proc_date = Time.now
  $osf.print "#times #{time_string}\n"
  if $clog_level > 0
    print "#Summary: times #{time_string}\n";  $stdout.flush
  end

  $osf.print "\n#eof \n"
  $osf.close
end  # write_sum_file()
