#!/usr/bin/env ruby

#  2359, Tue 29 Mar 11 (CET)
#
#  src_split.rb: Make separate .bpf files for list of src_addrs
#
#  Copyright 2011, Nevil Brownlee, U Auckland | CAIDA

require 'RubyLibtrace'
require './os-times'

class OutTraces
  def open_ott(source, dn)
    sp = IPprefix.from_s(source)
    ot = OutputTrace.new("pcapfile:#{dn}/#{source}.bpf")
    # ot.conf_compress_type(ZLIB_COMPRESSION)
    # ot.conf_compress_level(5)
    ot.start_output
    @ph[sp.addr] = ot
# print "sp= #{sp.version}, >#{sp.addr}<, #{sp.length}\n"
  end

  def initialize(tr_fn, s_fn, d_n, rg, n_sources)
    @pt = Trace.new("pcapfile:#{tr_fn}")
    @pt.start

    @ph = Hash.new
   
    check_dir_exists(d_n)
    f = File.open(s_fn, 'r')
    f.each do |line|
      aa = line.split
      if aa[0] == rg  # 4-digit string,
                  # e.g. 0503 -> XP_left, TCP horiz scan
        open_ott(aa[2], d_n)
        break
      end
    end
    ns = 1
    if ns != n_sources
      f.each do |line|
        aa = line.split
        open_ott(aa[2], d_n)
        ns += 1;  break if ns == n_sources
      end
    end
    f.close
  end

  def process
    np = 0
    @pt.each_packet do |pkt|
      s = pkt.ip.src_prefix
      ot = @ph[s.addr]
      if !ot.nil?
        ot.write_packet(pkt)
        np += 1  #;  break if np == 10
        print '.' if np % 10000 == 0
      end
    end

    @ph.each { |source, ot| ot.close_output }
    @pt.close
  end
end  # OutTraces

$st = OStimes.new
tfn = '/scratch/owtmon/net44.1299596400.pcap.gz'  # odin
# tfn = '/home/nevil/scratch/20110308-5m.bpf'  # nebbiolo

def run(tfn, which)
  $st.reset
  ott = OutTraces.new(tfn, 'sources-by-group.txt', which, which, 10)
  ott.process
  $st.session_times($stdout, which)
end

run(tfn, '0011')
run(tfn, '0009')

=begin
0011  # Other, TCP+UDP
0009  # Other, ICMP only

run(tfn, '0003')
run(tfn, '0503')
run(tfn, '0403')
run(tfn, '0003')
run(tfn, '0205')
run(tfn, '0008')

0003  # Other, TCP horizontal scan
0503  # XP_left, TCP horizontal scan
0403  # XP spkies, TCP horizontal scan
0003  # Other, TCP Horiz scan
0205  # DOS bursty, UDP probe
0008  # Other, UDP unknown
=end
